#!/usr/bin/python

# Primary Interface for the Sync Client

import os, sys, posixpath
import hashlib, base64, re, logging
import datetime, time

# Time-Drive Imports
from settings import appSettings, collectionSettings
from local import SyncFile, createDataConnection, createDataTables, \
	getLocalFileRecord, getLocalFolderRecord
from remote import RemoteResource, getRemoteCacheRecord
from syncxml import parseManifest
from syncaction import checkMd5, updateRemoteCache, \
	downloadNewFile, uploadNewFile, uploadUpdatedFile, \
	replaceLocalFile, missingManifestFile, \
	mergeExistingFile, deleteRemoteFile, deleteLocalFile, \
	downloadRemoteAsConflict, updateLocalFolderRecord_FromRemote, \
	updateLocalFolderRecord_FromLocal
from dateutils import *
from pathutils import *

import errors

# Class Enumerations
NEW = 0
UPDATE = 1
DELETE = 2
RENAME = 3
CHECK = 4

LOCAL_CHECK_NEW = 'Check-NewFile'
LOCAL_CHECK_EXISTING = 'Check-ExistingFile'
LOCAL_CHECK_DELETION = 'Check-DeletedFile'
LOCAL_CHECK_FOLDER_CHANGES = 'Check-Local-FolderChanges'

RT_SUCCESS = 'Op-Success'
RT_FAIL = 'Op-Fail'
RT_NOTNEEDED = 'Op-NotNeeded'

RT_UPLOAD_NEW = 'Upload-NewFile'
RT_UPLOAD_UPDATE = 'Upload-ExistingFile'
RT_DELETE_REMOTE = 'Delete-RemoteFile'
RT_DELETE_LOCAL = 'Delete-LocalFile'
RT_MOVE_REMOTE = 'Move-RemoteFile'
RT_MOVE_LOCAL = 'Move-LocalFile'

RT_CHECK_FOLDER_CHANGES = 'Check-Remote-FolderChanges'

# Sync Variables
FILE_CHUNK = 1024
OPS_PAUSE = 0.5

## File Sync Methods ##

def downloadCollection(wd_, server_url, 
		sync_backend, db_session, rd_filelist):
	'''The file collection doesn't yet exist on the local machine. Download
		all files to a local directory, set the appropriate permissions, create
		a sync database.'''

	logging.info('Starting Collection Download: ' + server_url)

	for file_ in rd_filelist:

		# Remote file data
		rdfile_ = rd_filelist[file_]

		# Try and retrieve local file data
		dbfile_ = getLocalFileRecord(rdfile_.local_path, db_session)

		# If no local file, download the file
		if (dbfile_ is None):
			status_, conflict_ = downloadNewFile(rdfile_, db_session, sync_backend)

		# Compare the two versions, download new files
		else:
			# For newer files, download and replace the local file
			if (rdfile_.mdate > dbfile_.mdate):
				status_, conflict_ = replaceLocalFile(
					dbfile_, rdfile_, db_session, sync_backend)
			elif (rdfile_.mdate == dbfile_.mdate) \
				& (rdfile_.md5_beg == dbfile_.md5_beg) \
				& (rdfile_.md5_end == dbfile_.md5_end):
				logging.info('File already at latest version: ' \
					+ rdfile_.resource_name)

	# Commit any unresolved changes
	db_session.commit()


def folderBackup(wd_, server_url, 
		sync_backend, db_session, rd_verified=False):
	'''Backup files in a particular folder. Requires the path to
		the working directory, the path to the server, an instantiated
		sync backend, and an open database session. Works recursively.'''

	logging.info('Starting Backup: ' + server_url)

	# Retrieve directory contents
	wd_contents = os.listdir(wd_)

	# Sync each file to cloud
	for file_ in wd_contents:

		# Full path to file
		file_fullpath = os.path.join(wd_, file_)

		if (os.path.isfile(file_fullpath) == True) \
			& (file_ not in sync_backend.exclude):

			# Retrive information about the file from the database
			dbfile_ = db_session.query(SyncFile).filter_by(
				full_path=file_fullpath)\
			.filter((SyncFile.file_deleted != True) | 
				(SyncFile.file_deleted == None))\
			.first()

			# Retrieve the date created and modified
			file_cdate = datetime.datetime.utcfromtimestamp(
				os.path.getctime(file_fullpath))
			file_mdate = datetime.datetime.utcfromtimestamp(
				os.path.getmtime(file_fullpath))

			# If there is no file, or the cdate/mdate are different
			# than those in the database, inspect the contents of the file
			if (dbfile_ is None) or (file_cdate is not dbfile_.cdate) \
				or (file_mdate is not dbfile_.mdate):
				hash_beg, hash_end = checkMd5(file_fullpath, FILE_CHUNK)

			# Determine the Destination URL of the File
			file_url = posixpath.join(server_url, file_)
			
			# New Files
			if dbfile_ is None:

				# Check to see if upload destination exists
				if rd_verified == False:
					rd_verified = sync_backend.checkFolder(
						server_url, create_folder=True)

				uploadNewFile(os.path.join(wd_, file_), # Local Path
					file_url, # Remote path
					(file_cdate, file_mdate), # File cdate and mdate
					(hash_beg, hash_end), # md5 for file beg and end
					db_session, sync_backend) # Database session and backend
			
			# Compare file to that found in database
			else:

				# File Unchanged, Do Nothing
				if (file_cdate is dbfile_.cdate) \
					and (file_mdate is dbfile_.mdate):
					pass
				if (hash_beg == dbfile_.md5_beg) and (hash_end == dbfile_.md5_end):
					pass

				# File modification date is different, Check md5
				elif (file_mdate > dbfile_.mdate):
					
					# md5 is the same, file unchanged
					if (hash_beg is dbfile_.md5_beg) and (hash_end is dbfile_.md5_end):
						dbfile_.setDateTimes(file_cdate, file_mdate) # Update times
						db_session.commit()
					
					# Upload a new version of the file to the server
					else:

						# Check to ensure the remote directory is verified
						if rd_verified == False:
							rd_verified = sync_backend.checkFolder(
								server_url, create_folder=True)
						
						uploadUpdatedFile(dbfile_, file_url, # dbfile, url
							(file_cdate, file_mdate), # cdate, mdate of file
							(hash_beg, hash_end), # md5 signatures
							db_session, sync_backend) # Database session and sync backend

		# Backup Files located in subfolders
		elif os.path.isdir(os.path.join(wd_, file_)) == True:
			subfolder_url = posixpath.join(server_url, file_)
			folderBackup(os.path.join(wd_, file_),
				subfolder_url,
				sync_backend,
				db_session)

	# Commit the changes to the database
	db_session.commit()


def updateCache(rd_filelist, db_session, sync_backend):
	'''Updates the cache of the remote files. Copies the file mdate/cdate,
		file signatures and other information needed to detect changes
		in the remote collection.'''

	# Iterate through the files in the remote collection
	for rdfile_localpath in rd_filelist.keys():

		# Retrieve the RemoteResource
		rdfile_ = rd_filelist.pop(rdfile_localpath)
		updateRemoteCache(rdfile_, db_session)

	db_session.commit()


def folderSync(collection_name, wd_, server_url, 
		sync_backend, db_session, rd_filelist,
		rd_verified = False, subfolder=False, manifest=False):
	'''Synchronize files in a particular directory. Similar to folderBackup,
		except that it will also download updated files from other machines
		and files that were uploaded through the web interface. '''

	sync_conflicts = []
	
	logging.info('Starting Sync: ' + server_url)

	# Retrieve directory contents, allow wd_ to drive sync
	wd_parent, wd_name = os.path.split(wd_)
	wd_contents = os.listdir(wd_)

	# Compare each file to database record and remote file
	for file_ in wd_contents:

		# Generate the full path to the file
		file_localpath = os.path.join(wd_, file_)
		if (os.path.isfile(file_localpath) == True) \
			& (file_ not in sync_backend.exclude):

			# Retrieve database and remote records about the file
			dbfile_ = db_session.query(SyncFile).filter_by(
				full_path=file_localpath)\
				.filter((SyncFile.file_deleted != True) | 
					(SyncFile.file_deleted == None))\
				.first()
			try: rdfile_ = rd_filelist.pop(file_localpath)
			except KeyError: rdfile_ = None

			# Remote file url
			file_url = posixpath.join(server_url, file_)

			# Retrieve information about the file on disk
			
			# File cdate
			try: file_cdate = datetime.datetime.utcfromtimestamp(
				os.path.getctime(file_localpath))
			except ValueError: file_cdate = datetime.datetime.utcnow()
			
			# File mdate
			try: file_mdate = datetime.datetime.utcfromtimestamp(
				os.path.getmtime(file_localpath))
			except ValueError: file_mdate = datetime.datetime.utcnow()

			# New file on the local disk
			if (dbfile_ is None) & (rdfile_ is None):

				# Retrieve file information
				md5_beg, md5_end = checkMd5(file_localpath, FILE_CHUNK)

				# Verify remote folder
				if rd_verified == False:
					rd_verified = sync_backend.checkFolder(
						server_url, create_folder = True)

				# Upload to the server
				result_, conflict_filepath = uploadNewFile(
					file_localpath, # Local Path
					file_url, # Remote path
					(file_cdate, file_mdate), # File cdate and mdate
					(md5_beg, md5_end), # md5 for file beg and end
					db_session, sync_backend) # Database session and backend
				if not result_:
					sync_conflicts.append(conflict_filepath)

			# Version of the file in the manifest and on local computer
			# check to see if they are the same. If so, add to local database.
			elif (dbfile_ is None) & (rdfile_ is not None):
				result_, conflict_filepath = mergeExistingFile(
					file_localpath, # Local Path
					file_url, # Remote Path
					(file_cdate, file_mdate), # File cdate and mdate
					rdfile_, # Remote File
					db_session, sync_backend)
				if not result_:
					sync_conflicts.append(conflict_filepath)

			# Version of file in database, but not in manifest
			# Check to see if file deleted, or just missing from incomplete sync
			elif (dbfile_ is not None) & (rdfile_ is None):

				# Check remote cache to see if there is a record of the file
				rcfile_ = getRemoteCacheRecord(dbfile_.full_path, db_session)
				
				if rcfile_ is not None:
					deleteLocalFile(dbfile_, rcfile_, db_session)
				else:
					md5_beg, md5_end = checkMd5(dbfile_.full_path, FILE_CHUNK)
					rcfile_ = RemoteResource(dbfile_.name, server_url)
					rcfile_.setMd5(md5_beg, md5_end)
					rcfile_.setDateTimes(dbfile_.cdate, dbfile_.mdate)
					deleteLocalFile(dbfile_, rcfile_, db_session)

			# Both the local and the remote file exist
			elif (dbfile_ is not None) & (rdfile_ is not None):

				# Newer file on server (database/local file mdate match, 
				# different md5 signatures)
				if (rdfile_.mdate > dbfile_.mdate) \
					& (dbfile_.mdate == file_mdate) \
					& ((rdfile_.md5_beg != dbfile_.md5_beg) \
						| (rdfile_.md5_end != dbfile_.md5_end)):

					result_, conflict_filepath = replaceLocalFile(
						dbfile_, # Database reference
						rdfile_, # Remote file reference
						db_session, sync_backend) # db session/sync backend
					if not result_:
						sync_conflicts.append(conflict_filepath)

				# Updated file on disk (database/remote file mdate match)
				elif (file_mdate > rdfile_.mdate) \
					& (dbfile_.mdate == rdfile_.mdate):

					# Retrieve information about the file
					md5_beg, md5_end = checkMd5(file_localpath, FILE_CHUNK)

					# Verify remote folder
					if rd_verified == False:
						rd_verified = sync_backend.checkFolder(
							server_url, create_folder = True)

					# Check file md5 against remote to ensure differences
					if (md5_beg != rdfile_.md5_beg) | (md5_end != rdfile_.md5_end):

						result_, conflict_filepath = uploadUpdatedFile(
							dbfile_, file_url, # database reference, remote url
							(file_cdate, file_mdate), # file cdate and mdate
							(md5_beg, md5_end), # file md5 checksums
							db_session, sync_backend) # db session/sync backend
						if not result_:
							sync_conflicts.append(conflict_filepath)

				# File on local disk is newer than server and remote
				# Remote is newer than that in database
				# Download the server file as conflict
				# Upload the local file as canonical
				elif (rdfile_.mdate > dbfile_.mdate) \
					& (file_mdate > dbfile_.mdate) \
					& (file_mdate > rdfile_.mdate):

					# Download remote file as conflict with timestamp
					result_, conflict_filepath = downloadRemoteAsConflict(
						rdfile_, file_localpath, sync_backend)
					logging.info('File Conflict Encountered: ' + file_)
					logging.info('Downloaded remote file to: ' + conflict_filepath)
					sync_conflicts.append(conflict_filepath)

					# Retrieve information about local file
					md5_beg, md5_end = checkMd5(file_localpath, FILE_CHUNK)

					# Upload local file as canonical
					result_, conflict_filepath = uploadUpdatedFile(
						dbfile_, file_url, # database reference, remote url
						(file_cdate, file_mdate),
						(md5_beg, md5_end),
						db_session, sync_backend)

		# Sync files in subfolders
		elif os.path.isdir(os.path.join(wd_, file_)):
			subfolder_url = posixpath.join(server_url, file_)
			subfolder_localpath = os.path.join(wd_, file_)

			# Check remote file list to make sure it
			# includes all remote files
			rd_filelist = sync_backend.remoteFileList(collection_name,
				subfolder_localpath, subfolder_url,
				rd_filelist)

			# Sync the subfolder
			folderSync(collection_name,
				os.path.join(wd_, file_),
				subfolder_url,
				sync_backend,
				db_session,
				rd_filelist,
				subfolder=True)

	# Add new files from the server, remove files that have been deleted
	# in the local collection, 
	if not subfolder:

		for rdfile_localpath in rd_filelist.keys():

			# Check to see if there is a record in the database
			dbfile_ = db_session.query(SyncFile).filter_by(
				full_path=rdfile_localpath)\
				.filter((SyncFile.file_deleted != True) | 
					(SyncFile.file_deleted == None))\
				.first()
			
			# New File - Download to the local disk
			if dbfile_ is None:
				rdfile_ = rd_filelist.pop(rdfile_localpath)
				if (rdfile_.md5_beg is None) & (rdfile_.md5_end is None):
					downloadNewFile(rdfile_, db_session, sync_backend, server_file=True)
				else: downloadNewFile(rdfile_, db_session, sync_backend)


			# Delete File - Old version on server
			elif dbfile_ is not None:
				rdfile_ = rd_filelist.pop(rdfile_localpath)
				# Check to see if the md5 checksums match
				if (rdfile_.md5_beg == dbfile_.md5_beg) \
					& (rdfile_.md5_end == dbfile_.md5_end):
					deleteRemoteFile(rdfile_, dbfile_, 
						db_session, sync_backend)

	# Commit any pending changes to the database
	db_session.commit()


def localFolderChanges(collection_, wd_, db_session,
	change_list = None, parent_dbfolder = None):
	'''Analyze the contents of a local folder on disk and determine what has changed.'''

	# If no changelist supplied, create dictionary to keep track of modified files
	if change_list == None: changelist_ = []
	else: changelist_ = change_list
	folder_modified = 0

	# Retrieve reference to local
	top_dbfolder = getLocalFolderRecord(wd_, db_session)

	# Retrieve mdate and cdate for the directory
	top_dbfolder_cdate, top_dbfolder_mdate = getFileDates(wd_)

	# Compare database mdate with database mdate
	if top_dbfolder is not None:

		# If there is a match, there are no changes to the directory
		if (top_dbfolder_mdate == top_dbfolder.mdate):
			logging.info('Top-level mdate matches for collection: ' \
				+ str(collection_['Collection-Name']))
			return changelist_

	# Compare files and folders, collect files that have been modified

	# Retrieve list of the folder contents
	try: wd_contents = os.listdir(wd_)
	except OSError: wd_contents = []

	# Iterate through contents of directory, compare file information to database
	for resource_localname in wd_contents:
		resource_localpath = os.path.join(wd_, resource_localname)
		resource_cdate, resource_mdate = getFileDates(resource_localpath)

		# Compare folders in the directory to those in the database
		if os.path.isdir(resource_localpath):
			analyze_folder = True
			dbfolder_ = getLocalFolderRecord(resource_localpath, db_session)

			if dbfolder_:
				if resource_mdate == dbfolder_.mdate: analyze_folder = False

			# Analyze subfolders where the mdate values differ
			if analyze_folder:
				logging.info('Looking for local changes in ' + resource_localpath)
				changelist_ = localFolderChanges(collection_, resource_localpath, 
					db_session, change_list = changelist_, parent_dbfolder=dbfolder_)

		# Compare files in the directory to those in the database
		# Compares both the mdate and the md5 signature to detect files
		else:
			dbfile_ = getLocalFileRecord(resource_localpath, db_session)

			if dbfile_:

				# Compare mdate values
				if resource_mdate > dbfile_.mdate:

					# Compare Md5
					md5_beg, md5_end = checkMd5(resource_localpath, FILE_CHUNK)
					if (md5_beg != dbfile_.md5_beg) | (md5_end != dbfile_.md5_end):
						changelist_.append(resource_localpath)
						folder_modified += 1

			# Unable to find database record, add to changelist
			else:
				changelist_.append(resource_localpath)
				folder_modified += 1

	# Update local folder records (parent folder)
	if (parent_dbfolder is not None) & (folder_modified == 0):
		logging.info('Updating folder database values: ' + parent_dbfolder.name)
		parent_cdate, parent_mdate = getFileDates(parent_dbfolder.full_path)
		updateLocalFolderRecord_FromLocal(parent_dbfolder.full_path, 
			(parent_cdate, parent_mdate),
			parent_dbfolder, db_session)

	# Update top level folder record
	if (parent_dbfolder == None):
		if len(changelist_) == 0:
			logging.info('Top-Level Folder Record (Local): ' + wd_)
			updateLocalFolderRecord_FromLocal(
				wd_, (top_dbfolder_cdate, top_dbfolder_mdate),
				top_dbfolder, db_session)

	# Save any pending changes to the database
	db_session.commit()

	return changelist_


def cleanDatabase(wd_, db_session):
	'''Retrieve a list of files for the current working directory. Compare
		the files on disk to those in the database. If there are extra
		files in the directory, delete them from disk.'''

	logging.info('Analyzing Directory: ' + wd_)

	# Retrieve a list of files for the directory

	wd_contents = os.listdir(wd_)
	for file_ in wd_contents:

		file_localpath = os.path.join(wd_, file_)
		
		if os.path.isfile(file_localpath):

			# Retrieve database entries for the file
			db_files = db_session.query(SyncFile)\
				.filter_by(full_path=file_localpath)\
				.filter((SyncFile.file_deleted != True) | 
					(SyncFile.file_deleted == None))\
				.order_by(SyncFile.mdate).all()
			
			# If more than 1 active entry, mark later entries as deleted
			if len(db_files) > 1:
				logging.info('Removing duplicate file entries: ' + file_localpath)
				for dbfile_ in db_files[1:]:
					dbfile_.file_deleted = True
					dbfile_.ddate = datetime.datetime.utcnow()
					
			# Verify data
			if len(db_files) > 0:
				dbfile_ = db_files[0]
				if dbfile_.file_deleted is not True:
					# File name and directory
					if dbfile_.name != str(file_):
						logging.info('Fixing incorrect file name: ' + file_localpath)
						dbfile_.name = file_
					if dbfile_.parent_directory != str(wd_):
						logging.info('Fixing incorrect directory: ' + file_localpath)
						dbfile_.parent_directory = wd_

		# Process subdirectories
		elif os.path.isdir(file_localpath):
			cleanDatabase(file_localpath, db_session)

	# Remove old records and records with a valid ddate
	db_files_deleted = db_session.query(SyncFile)\
		.filter_by(parent_directory=wd_)\
		.filter(SyncFile.file_deleted == True).all()
	for deleted_dbfile in db_files_deleted:
		if deleted_dbfile.ddate is None:
			db_session.delete(deleted_dbfile)
		elif type(deleted_dbfile.ddate) is datetime.datetime:
			diff_ = datetime.datetime.utcnow() - deleted_dbfile.ddate
			if diff_ > datetime.timedelta(days=180):
				db_session.delete(deleted_dbfile)

	# Save pending changes to the folder
	db_session.commit()

## Manifests/Records ##

def createFullSyncManifest(collection_name, collection_wd, db_session):
	'''Create an XML log of all files in the database and upload it to 
		the remote server. Used by other sync clients to determine 
		if their file collections are up to date or if they have files
		which need to be downloaded and added to the local copy.'''

	# Create Sync Manifest Structure
	sync_time = datetime.datetime.utcnow()
	sync_root, sync_prop, sync_contents = syncxml.syncManifest(
		collection_name, sync_time)

	# Retrieve a collection of all files in the database
	db_syncfiles = db_session.query(SyncFile)\
		.filter((SyncFile.file_deleted != True) | 
			(SyncFile.file_deleted == None))\
		.order_by(SyncFile.full_path).all()
	for dbfile_ in db_syncfiles:
		sync_contents.append(dbfile_.xmlEntry(collection_wd))
	return sync_root
