#!/usr/bin/python

import posixpath, datetime, os, urlparse
from backends import dav, owncloud
import errors
from lxml import etree

BACKEND_PROVIDERS = {
	'dav' : dav, 'owncloud' : owncloud,
}

class SyncBackend:
	'''Class used for communicating with remote servers'''

	server_url = None
	server_user = None
	server_password = None
	povider_type = None
	provider = None
	exclude = None
	settings_url = None

	def __init__(self, server_url, server_user, server_password, backend_type):
		'''Initialize the backend, determine which provider should be used'''
		self.server_url = server_url
		self.server_user = server_user
		self.server_password = server_password
		self.exclude = []

		# Check to make sure backend is available
		if backend_type in BACKEND_PROVIDERS.keys():			
			self.provider_type = backend_type
			self.provider = BACKEND_PROVIDERS[backend_type]

	def setExcludeList(self, exclude_list):
		'''Set the exclude list for the backend'''
		self.exclude = exclude_list

	def settingsFolder(self, settings_folder):
		'''Folder to be used for Time Drive settings and archive'''
		self.settings_url = posixpath.join(self.server_url, settings_folder)

	def checkFile(self, file_url):
		'''Check to see if a remote file exists'''

		result_ = self.provider.checkWebFile(file_url,
			self.server_user, self.server_password)
		return result_

	def checkFolder(self, folder_url, create_folder=False):
		'''Check to make sure that the web folder exists'''

		result_ = self.provider.checkWebFolder(folder_url, 
				self.server_user, self.server_password)

		# Authentication or network error
		if result_ == None: return None
		
		# If create_folder is set to true, create the folder
		# and then query a second time
		if (create_folder is True) and (result_ is False):

			# Start by checking the parent folder
			parfolder_result = self.checkFolder(
				posixpath.dirname(folder_url), create_folder=True)
			
			# If there is a valid parent folder, create the child
			if parfolder_result:
				self.createFolder(folder_url)
				result_ = self.provider.checkWebFolder(folder_url,
					self.server_user, self.server_password)

		return result_

	def createFolder(self, folder_url):
		'''Create a folder at the destination'''
		self.provider.createWebFolder(folder_url,
			self.server_user, self.server_password)

	def folderContents(self, folder_url):
		'''Retrieve a list of the contents of a particular folder'''
		folder_contents = self.provider.folderContents(
			folder_url, self.server_user, self.server_password)
		return folder_contents

	def checkLocalFolder(self, local_folder, create_folder=False):
		'''Check to make sure that the local folder exists'''

		result_ = os.path.exists(local_folder)
		
		if (result_ == False) & (create_folder == True):

			# Check the parent folder
			parfolder_result = self.checkLocalFolder(
				os.path.dirname(local_folder), create_folder == True)
			
			# Create the directory, then check a second time
			if parfolder_result:
				os.mkdir(local_folder)
				result_ = os.path.exists(local_folder)

		return result_

	def uploadFile(self, file_url, file_contents, 
			file_mime, file_dates, localfile_path=None, file_md5=None):
		'''Upload file to the specified url'''
		r_ = self.provider.uploadFile(file_url, 
			self.server_user, self.server_password,
			file_contents, file_mime, file_dates, 
			localfile_path, file_md5)
		return r_

	def updateProperties(self, file_url, property_list):
		'''Send a property update for a particular resource. Requires the
			file_url and a property list. The property list is a dictionary
			of property names and their values.'''
		
		fileurl_ = self.remoteFileUrl(file_url)
		r_ = self.provider.updateProperties(fileurl_,
			self.server_user, self.server_password, property_list)
		return r_

	def deleteFile(self, file_url):
		'''Delete the file at the specified url'''
		fileurl_ = self.remoteFileUrl(file_url)
		r_ = self.provider.deleteFile(fileurl_,
			self.server_user, self.server_password)
		return r_

	def remoteFileUrl(self, file_url):
		'''Check the supplied URL, if malformed, supplement with
			download details from the backend.'''

		# Check url to ensure that it is well formed
		# If it is, return the url as valid
		valid_url = None
		url_file = urlparse.urlparse(file_url)
		if (url_file.scheme is not '') & (url_file.netloc is not '') \
			& (url_file.path is not ''):
			valid_url = file_url

		# If not, fill in additional details from the backend
		else:
			url_server = urlparse.urlparse(self.server_url)
			server_ = url_server.scheme + '://' + url_server.netloc
			fileurl_ = posixpath.join(server_, url_file.path.lstrip('/'))
			valid_url = fileurl_

		return valid_url

	def downloadFile(self, file_url):
		'''Download a file from the specified url'''

		r_ = None

		fileurl_ = self.remoteFileUrl(file_url)
		r_ = self.provider.downloadFile(fileurl_, 
			self.server_user, self.server_password)
		
		return r_

	def uploadManifest(self, collection_, manifest_):
		'''Upload the manifest file to the Time Drive folder on the server'''

		# Check Upload Destination
		self.checkFolder(posixpath.join(
			self.settings_url, 'manifest'), 
			create_folder=True)

		# Manifest Date/Times, Name, Url
		manifest_cdate = datetime.datetime.utcnow()
		manifest_filename = collection_ + '.' + 'full' + '.' \
			+ manifest_cdate.strftime('%Y-%m%d.%H%M') + '.manifest'
		manifest_url = posixpath.join(
			self.settings_url, 'manifest', manifest_filename)
		manifest_mime = 'text/xml'
		self.uploadFile(manifest_url, manifest_, 
			manifest_mime, (manifest_cdate, manifest_cdate))

	def downloadManifest(self, collection_):
		'''Download the most recent manifest for the collection.'''
		return self.provider.downloadManifest(collection_, self)

	def remoteFileList(self, collection_, wd_, rd_, 
			rd_filelist = None, include_folders=False):
		'''Fetch the remote file list for a particular collection'''
		return self.provider.remoteFileList(collection_, self, wd_, rd_, 
			rd_filelist, include_folders)

	def remoteFileDetails(self, rd_):
		'''Fetch the details of a specific file. Returns a remote file object.'''
		valid_url = self.remoteFileUrl(rd_)
		return self.provider.remoteFileDetails(self, valid_url)

	def finishSync(self, collection_, wd_, db_session):
		'''Perform final actions for the backend'''
		self.provider.finishSync(collection_, wd_, self, db_session)

	def remoteFolderChanges(self, collection_, rd_, wd_, db_session):
		return self.provider.remoteFolderChanges(collection_, self, 
			rd_, wd_, db_session)