#!/usr/bin/python

import logging, time, base64, datetime
from lxml import etree
import os, requests, posixpath
import urllib

# Time Drive Imports
import sync
import sync.syncxml as syncxml
from sync.remote import RemoteResource

## WebDav Related Structures

dav_prefixes = {'D' : 'DAV:', 'td' : 'http://www.oak-tree.us/time-drive/',
	's' : 'http://sabredav.org/ns'}
dav_default_properties = {'getetag' : 'D', 
	'getlastmodified' : 'D', 'resourcetype' : 'D', 
	'getcontenttype' : 'D', 'getcontentlength' : 'D'}
dav_sync_properties = {'md5beg' : 'td', 'md5end' : 'td'}

## Exceptions and Errors
sabredav_exceptions = {
	'Sabre_DAV_Exception_NotAuthenticated' : sync.errors.ERROR_AUTH,
	'Sabre_DAV_Exception_NotFound' : sync.errors.ERROR_NOTFOUND
}

# Status Codes
proppatch_success = [200, 207]

# Backend Properties
RT_Sync = False

## Generate WebDav XML ##

def createPropFind(dav_custom_properties = []):
	'''Generate XML request for webdav PROPFIND.'''

	# Create the Root Request for the Object
	propfind = etree.Element('{'+dav_prefixes['D']+'}propfind', xmlns='DAV:', nsmap = dav_prefixes)
	prop_ = etree.SubElement(propfind, '{'+dav_prefixes['D']+'}prop')
	
	# For requests that involve custom properties
	if len(dav_custom_properties) > 0:

		# Add Default Properties to the Request
		for propname_ in dav_default_properties.keys():
			prop_xml = etree.SubElement(prop_, 
				'{'+dav_prefixes[dav_default_properties[propname_]]+'}'+propname_)

		# Add custom properties
		for custom_propname in dav_custom_properties:
			prop_xml = etree.SubElement(prop_, custom_propname)
	
	return propfind

def createPropPatch(property_list):
	'''Generate a prop-patch which can be used to update file properties,
		such as the date/time that a file was last modified'''

	proppatch_ = etree.Element('{'+dav_prefixes['D']+'}propertyupdate', 
		nsmap = dav_prefixes)
	proppatch_set = etree.SubElement(proppatch_, '{'+dav_prefixes['D']+'}set')
	proppatch_prop = etree.SubElement(proppatch_set, '{'+dav_prefixes['D']+'}prop')
	
	# Iterate through property list
	for property_ in property_list.keys():
		proppatch_propname = etree.SubElement(proppatch_prop, property_)
		proppatch_propname.text = property_list[property_]
	return proppatch_


## Send WebDav Requests ## 

def sendMkCol(url_, user_, password_, request_headers, r_result=0):
	'''Send a MKCOL request, retry if there is a connection error'''

	# Try to send the information
	try:
		r_ = requests.request('MKCOL', url_,
			auth = requests.auth.HTTPBasicAuth(user_, password_),
			headers=request_headers, verify = False)
		return r_, True

	# ConnectionError
	except requests.ConnectionError, e:
		if r_result < 3:
			logging.warning('Unable to make connection to ' + url_)
			logging.warning(e.message)
			logging.warning('Retrying connection (' + str(r_result+1) + ' of 3)')
			r_, r_status = sendMkCol(url, user_, password_,
				request_headers, r_result = r_result+1)
			return r_, True
		else:
			logging.warning('Connection Failed (MkCol): ' + url_)
			return None, False

def sendDeleteRequest(url_, user_, password_, request_headers, r_result=0):
	'''Send a DELETE request, retry if there is a connection error'''

	#Try to send the request
	try:
		r_ = requests.request('DELETE', url_,
			auth = requests.auth.HTTPBasicAuth(user_, password_),
			headers=request_headers, verify = False)
		return r_, True

	# ConnectionError
	except requests.ConnectionError, e:
		if r_result < 3:
			logging.warning('Unable to make connection to ' + url_)
			logging.warning(e.message)
			logging.warning('Retrying connection (' + str(r_result+1) + ' of 3)')
			r_, r_status = sendDeleteRequest(url_, user_, password_, 
				request_headers, r_result = r_result+1)
			return r_, True
		else:
			logging.warning('Connection Failed (Delete Item): ' + url_)
			return None, False

def sendPropFind(url_, user_, password_, request_headers, request_data, r_result=0):
	'''Send a PROPFIND request, retry if there is a connection error'''

	# Try to send the request
	try:
		r_ = requests.request('PROPFIND', url_,
			auth = requests.auth.HTTPBasicAuth(user_, password_),
			headers = request_headers, data = request_data, verify = False)
		return r_, True

	# ConnectionError
	except requests.ConnectionError, e:
		if r_result < 3:

			# Log Connection Failure, Retry Connection
			logging.warning('Unable to make connection to ' + url_)
			logging.warning(e.message)
			logging.warning('Retrying connection (' + str(r_result+1) + ' of 3)')
			r_, r_status = sendPropFind(url_, user_, password_,
				request_headers, request_data, r_result = r_result+1)
			return r_, True

		else:
			logging.warning('Connection Failed (PropFind): ' + url_)
			return None, False

	except: return None, False


def sendPropPatch(url_, user_, password_, request_headers, request_data, r_result=0):
	'''Send a PROPFIND request, retry if there is a connection error'''
	
	# Try to send the request
	try:
		r_ = requests.request('PROPPATCH', url_,
			auth = requests.auth.HTTPBasicAuth(user_, password_),
			headers = request_headers, data = request_data, verify = False)
		return r_, True

	# ConnectionError
	except requests.ConnectionError, e:
		if r_result < 3:
			logging.warning('Unable to make connection to ' + url_)
			logging.warning(e.message)
			logging.warning('Retrying connection (' + str(r_result+1) + ' of 3)')
			r_, r_status = sendPropPatch(url_, user_, password_,
				request_headers, request_data, r_result = r_result+1)
			return r_, True
		else:
			logging.warning('Connection Failed (PropPatch): ' + url_)
			return None, False

def sendFileUpload(url_, user_, password_, request_headers, 
		request_data, localfile_url=None, r_result=0):
	'''Use a PUT request to send a file to the web server. 
		Retry if there is a connection error.'''

	# Try to send the request
	try:

		if localfile_url:
			# Create a files data structure
			logging.debug('Loading data from file: ' + localfile_url)
			file_dir, file_name = posixpath.split(url_)
			files_ = {file_name : open(localfile_url, 'rb')}
			# Send Requests
			r_ = requests.put(url_, 
				auth = requests.auth.HTTPBasicAuth(user_, password_),
				headers = request_headers, 
				files = files_,
				verify = False)
			return r_, True

		elif request_data:
			logging.debug('Uploading file with generated data: ' + url_)
			r_ = requests.put(url_,
				auth = requests.auth.HTTPBasicAuth(user_, password_),
				headers = request_headers,
				data = request_data,
				verify = False)
			return r_, True

		else:
			logging.warning('No file path or data provided for ' + url_)
	
	# ConnectionError
	except requests.ConnectionError, e:
		if r_result < 3:
			logging.warning('Unable to make connection to ' + url_)
			logging.warning(e.message)
			logging.warning('Retrying connection (' + str(r_result+1) + ' of 3)')
			r_, r_status = sendFileUpload(url_, user_, password_,
				request_headers, request_data, localfile_url, r_result = r_result+1)
			return r_, True
		else:
			logging.warning('Connection Failed (UploadFile): ' + url_)
			return None, False

def sendFileDownload(url_, user_, password_, r_result=0):
	'''Use a GET request to download files from the web server.
		Retry if there is a connection error'''

	try:
		r_ = requests.get(url_, 
			auth=requests.auth.HTTPBasicAuth(user_, password_), 
			verify = False)
		return r_, True
	except:
		if r_result < 3:
			logging.warning('Unable to make connection to' + url_)
			logging.warning('Retrying connection (' + str(r_result+1) + ' of 3)')
			r_, r_status = sendFileDownload(url_, user_, password_,
				r_result = r_result+1)
			return r_, True
		else:
			logging.warning('Connection Failed (DownloadFile): ' + url_)
			return None, False


## Methods to Create/Modify/List Contents of WebDav Collections ##

def davFolderContents(url_, user_, password_, depth_):
	'''Retrieve the contents of a WebDav collection. Requires the url of the
		collection, the username and the password. Depth can be used to control
		how much XML is returned. A depth of 0 returns properties for only
		current object (used to verify that it exists), a depth of 1 returns
		properties for objects in the collection, but ignores objects in subfolders.
		A depth of infinity (or no depth at all), returns properties for all child 
		objects in the collection.'''

	# Create Request to Check Whether a Folder Exists
	custom_properties = []
	for dav_propname in dav_sync_properties.keys():
		clark_notation = '{'+dav_prefixes[dav_sync_properties[dav_propname]]+'}'+dav_propname
		custom_properties.append(clark_notation)

	request_data = etree.tostring(createPropFind(custom_properties), pretty_print=True,
		xml_declaration=True, encoding='UTF-8')
	request_headers = {'Depth' : str(depth_), 
		'Content-Type' : 'text/xml; charset="utf-8"',
		'Content-Length' : str(len(request_data))}

	#  Send Request
	r_, r_status =  sendPropFind(url_, user_, password_, 
		request_headers, request_data)

	# Return Results
	if (r_ is None): raise sync.errors.ConnectionError(sync.errors.ERROR_NETWORK)
	elif (r_status is True) & (r_ is not None): return r_.text
	else: return None

def createWebFolder(url_, user_, password_):
	'''Generate XML request for webdav mkcol and send. Returns the request status
		and the response from the server.'''
	logging.debug('Create web folder: ' + url_)

	request_headers = {'Content-Type' : 'text/xml',
		'Content-Length' : '0'}
	r_, r_status = sendMkCol(url_, user_, password_, request_headers)
	if r_status:
		return r_.status_code, r_.text
	else:
		return 404, ''

def checkWebFolder(url_, user_, password_):
	'''Check whether a particular folder exists. Returns true if it exists,
		returns false if the connection is valid, but unable to find a folder.
		Returns None if the connection is invalid.'''
	
	# Retrieve folder listing/properties
	logging.debug('Check web folder: ' + url_)
	try: folder_xml = davFolderContents(url_, user_, password_, 0)
	except sync.errors.ConnectionError: return None

	# Check to see if the request returned an object
	if folder_xml:

		# Parse Folder Request
		try: resource_url, resource_type, resource_modified, resource_status = \
			parseFolderCheck(folder_xml)
		
		# Manage errors raised during check
		except sync.errors.SyncError, e: 

			# Unable to find resource
			if e.type in [sync.errors.ERROR_NOTFOUND]: return False
			# For other errors, return None
			else: return None

		if (resource_type == 'collection') & (resource_status == '200'):
			return True

	# If unable to make connection, return None
	else: return None

def checkWebFile(url_, user_, password_):
	'''Check whether a particular file exists'''

	# Retrieve file properties
	logging.debug('Checking web file: ' + url_)
	file_xml = davFolderContents(url_, user_, password_, 0)
	if file_xml:
		resource_url, resource_type, resource_modified, resource_status = \
			parseFolderCheck(file_xml)
		if (resource_status == '200'): return True
		else: return False
	
	# If no object, return False
	return False

def folderContents(url_, user_, password_):
	'''Retrieve the contents of a particular folder'''

	# Retrieve listing of folder properties
	logging.debug('Fetching Folder Contents: ' + url_)
	folder_xml = davFolderContents(url_, user_, password_, 1)

	# If there is valid listing, parse the results
	if folder_xml:
		try: 
			folder_contents = parseFolderContentsRequest(folder_xml)
			return folder_contents
		except sync.errors.SyncError, e:
			if e.type == sync.errors.ERROR_NOTFOUND: return None

	# No listings, return null object
	else: return None

def parseFolderContentsRequest(folder_xml):
	'''Parse the results of a folder contents request. Return a list of RemoteFile
		objects corresponding to the contents of the directory.'''

	folder_contents = []
	
	# Try to parse the xml file
	try:
		xml_ = etree.XML(syncxml.removeXmlDeclarations(folder_xml))

	# Upon error, notify the user and return null objects
	except:
		logging.warning('Unable to def folderContentsparse XML file')
		logging.warning(folder_xml)
		return None

	if xml_.tag == '{'+dav_prefixes['D']+'}error':
		parseRemoteError(xml_)

	for node_ in xml_:
		if node_.tag == '{'+dav_prefixes['D']+'}response':
			resource_ = parseRemoteResource(node_)
			if resource_: folder_contents.append(resource_)

	return folder_contents

def parseRemoteResource(resource_xml):
	'''Parse resource responses from the server, returns a RemoteFile object'''

	resource_url = None
	resource_mdate = None
	resource_type = None
	resource_status = None
	resource_id = None
	sync_md5beg = None
	sync_md5end = None

	# Retrieve Resource Url
	xml_url = resource_xml.find('{'+dav_prefixes['D']+'}href')
	
	if xml_url is not None:
		resource_url = urllib.unquote(xml_url.text)

	# Retrieve Resource Properties
	xml_propstat = resource_xml.find('{'+dav_prefixes['D']+'}propstat')
	if xml_propstat is not None:
		xml_prop = xml_propstat.find('{'+dav_prefixes['D']+'}prop')
		xml_status = xml_propstat.find('{'+dav_prefixes['D']+'}status')
		
		if xml_prop is not None:
			xml_mdate = xml_prop.find('{'+dav_prefixes['D']+'}getlastmodified')
			xml_resourcetype = xml_prop.find('{'+dav_prefixes['D']+'}resourcetype')
			xml_retag = xml_prop.find('{'+dav_prefixes['D']+'}getetag')
			xml_md5beg = xml_prop.find('{'+dav_prefixes['td']+'}md5beg')
			xml_md5end = xml_prop.find('{'+dav_prefixes['td']+'}md5end')

			# Date Modified
			if xml_mdate is not None:
				resource_mdate = sync.string2datetime(xml_mdate.text)

			# Resource Type
			if xml_resourcetype is not None:
				for node_ in xml_resourcetype:
					resource_type = syncxml.trimNamespace(node_.tag)

			# Resource eTag
			if xml_retag is not None:
				resource_id = xml_retag.text

			# MD5 Values Set by the Client
			if xml_md5beg is not None:
				sync_md5beg = xml_md5beg.text
			if xml_md5end is not None:
				sync_md5end = xml_md5end.text

		# Resource Status
		if xml_status is not None:
			resource_status = syncxml.findHtmlStatus(xml_status.text)

	# Create Remote Resource
	if resource_url:
		resource_dir, resource_name = posixpath.split(resource_url)
		resource_ = RemoteResource(resource_name, resource_dir)
		resource_.setDateTimes(resource_mdate, resource_mdate)
		resource_.setResourceType(resource_type)
		resource_.setResourceStatus(resource_status)
		resource_.setEtag(resource_id)
		resource_.setMd5(sync_md5beg, sync_md5end)

	else:
		resource_ = None

	return resource_

def parseRemoteError(error_xml):
	'''Parse the error XML sent from the server, return reason for error'''

	# Determine the type of error
	xml_exception = error_xml.find('{'+dav_prefixes['s']+'}exception')
	xml_message = error_xml.find('{'+dav_prefixes['s']+'}message')
	if (xml_exception is not None) & (xml_message is not None):
		try: error_code = sabredav_exceptions[xml_exception.text]
		except: error_code = sync.errors.ERROR_SYNC
		if error_code == sync.errors.ERROR_AUTH:
			raise sync.errors.AuthenticationError(error_code, xml_message.text)
		else: raise sync.errors.SyncError(error_code, xml_message.text)
	else: raise sync.errors.SyncError(sync.errors.ERROR_SYNC)

def parseFolderCheck(folder_xml):
	'''Parse the XML from the server, return basic folder properties'''

	# Try to parse the xml file
	try: xml_ = etree.XML(syncxml.removeXmlDeclarations(folder_xml))

	# Upon error, notify the user and return null objects
	except:
		logging.warning('Unable to parse XML file')
		logging.warning(folder_xml)
		return None, None, None, None

	resource_ = None
	resource_url = None
	resource_type = None
	resource_modified = None
	resource_status = None

	# Check for errors in response
	if xml_.tag == '{'+dav_prefixes['D']+'}error':
		parseRemoteError(xml_)

	# Iterate through response nodes, parse for collection details
	for node_ in xml_:
		if node_.tag == '{'+dav_prefixes['D']+'}response':
			resource_ = parseRemoteResource(node_)

	if resource_:
		resource_url = resource_.full_path
		resource_type = resource_.resource_type
		resource_modified = resource_.mdate
		resource_status = resource_.resource_status

	return resource_url, resource_type, resource_modified, resource_status


# File Upload/Download Methods

def deleteFile(url_, user_, password_):
	'''Delete a file from a WebDav server using a delete request. Returns
		True if the upload was successful. Otherwise, returns False.'''

	logging.debug('Delete file from ' + url_)
	request_headers = {'Content-Length' : str(0)}
	r_, r_status = sendDeleteRequest(url_, user_, password_, request_headers)

	if r_status:
		if r_.status_code == 204: return True
	
	return False

def uploadFile(url_, user_, password_, 
		contents_, mime_, dates_, localfile_url=None, file_md5=None):
	'''Upload a file to a WebDav server using a put request. Returns True
		if the request was successful, otherwise returns False.'''
	
	logging.debug('Upload file to ' + url_)

	# Ensure that there is data to send
	contents_size = len(contents_)
	if contents_size > 0:

		# Extract Creation and Modified Dates
		cdate_ = dates_[0]
		mdate_ = dates_[1]

		# Create DAV Request Headers
		request_headers = {'Content-Type' : mime_}
		if contents_:
			request_headers['Content-Length'] = str(len(contents_))
		elif localfile_url:
			request_headers['Content-Length'] = str(os.path.getsize(localfile_url))

		# Send Request
		r_, r_status = sendFileUpload(url_, user_, password_, 
			request_headers, contents_, localfile_url)

		# Create Property Array
		prop_update = {'{'+dav_prefixes['D']+'}lastmodified' : str(int(sync.epochSeconds(mdate_)))}
		
		# Add md5 checksums to property list
		if file_md5:
			md5beg_, md5end_ = file_md5
			prop_update['{'+dav_prefixes['td']+'}md5beg'] = md5beg_
			prop_update['{'+dav_prefixes['td']+'}md5end'] = md5end_
		
		# Send Property Update
		updateDavProperties(url_, user_, password_, prop_update)

		if r_status:
			if r_.status_code == 201: return True
			else: return False
		else:
			return False

	# Nothing to upload
	return None

def updateProperties(url_, user_, password_, proplist_):
	'''Create a valid property update and send it to the server. Requires
		a property list as a dictionary with property names and values as
		key-pairs.'''

	property_list = {}

	# Add namespaces to the properties and values
	for propname_ in proplist_.keys():
		property_name = '{'+dav_prefixes['td']+'}'+propname_
		property_value = proplist_[propname_]
		property_list[property_name] = property_value

	r_proppatch = updateDavProperties(url_, user_, password_, property_list)

	if r_proppatch.status_code in proppatch_success: return True
	else: return False

def updateDavProperties(url_, user_, password_, properties_):
	'''Create and send an update request for a particular file'''

	proppatch_data = etree.tostring(
		createPropPatch(properties_),
		pretty_print=True,
		xml_declaration=True, encoding='UTF-8')
	proppatch_headers = {'Content-Type' : 'text/xml; charset="utf-8"',
		'Content-Length' : str(len(proppatch_data))}
	r_proppatch, r_proppatch_status = sendPropPatch(url_, user_, password_,
		proppatch_headers, proppatch_data)

	return r_proppatch

def downloadFile(url_, user_, password_):
	'''Download a file from the remote WebDav server. Returns the contents of
		the file if the request was successful, otherwise returns None.'''

	# Send request for file
	r_, r_status = sendFileDownload(url_, user_, password_)

	# Check request status and return results
	if r_status:
		return r_.content
	else:
		return None


# Methods for Working with the Manifest and the Remote File List

def remoteFileList(collection_, backend_, wd_, rd_, rdlist_ = None):
	'''Generate a list of remote files and their properties which can be used
		in synchronizing the remote and local collections.'''

	if rdlist_ is None:

		# Download the Xml Manifest for the Collection
		manifest_xml = downloadManifest(collection_, backend_)
		if manifest_xml:
			rd_filelist = sync.parseManifest(manifest_xml, wd_, rd_)
		if not rd_filelist:
			rd_filelist = {}

	else: rd_filelist = rdlist_

	# Merge the list of files in the current remote directory
	# with those of the manifest. This will detect new uploads
	# that have not yet been synced.
	server_contentsraw = sync_backend.folderContents(server_url)
	try: server_filelist = dict(
		(os.path.join(wd_, f_.resource_name), f_) for f_ in server_contentsraw)
	except: server_filelist = {}

	# Retrieve the webdav entry for the remote working directory
	if sys.platform == 'win32':
		try: wd_daventry = server_filelist.pop(wd_+'\\')
		except: wd_daventry = None
	else: wd_daventry = server_filelist.pop(wd_+'/')

	# Check to see if the file is already in the manifest, if not, add it
	for rdfile_filename in server_filelist.keys():
		if rdfile_filename not in rd_filelist.keys():
			rd_filelist[rdfile_filename] = server_filelist[rdfile_filename]

	return rd_filelist

def downloadManifest(collection_, backend_):
	'''Download the most recent manifest for the collection'''

	collection_manifests = []
	manifest_mostrecent = None
	manifest_mostrecent_datetime = datetime.datetime(year=1970, month=1, day=1)

	# Retreive the most recent manifest for the collection
	log_url = posixpath.join(backend_.settings_url, 'manifest')
	folder_contents = folderContents(log_url, 
		backend_.server_user, backend_.server_password)

	if folder_contents:

		# Find the most recent manifest for the collection
		for file_ in folder_contents:
			if file_.resource_name.startswith(collection_):
				collection_manifests.append(file_)
		for manifest_ in collection_manifests:
			manifest_collection, manifest_type, manifest_date, \
				manifest_time, manifest_ext = manifest_.resource_name.split('.')
			manifest_datetime = datetime.datetime.strptime(
				'.'.join([manifest_date, manifest_time]), '%Y-%m%d.%H%M')
			if manifest_datetime > manifest_mostrecent_datetime:
				manifest_mostrecent = manifest_
				manifest_mostrecent_datetime = manifest_datetime

		if manifest_mostrecent:
			manifest_download = manifest_mostrecent.downloadFile(backend_)
			if manifest_download:
				# Open Temp File from Disk
				manifest_file = open(manifest_mostrecent.resource_contents, 'r+b')
				manifest_contents = manifest_file.read()
				manifest_file.close()
				# Delete Temp File
				manifest_mostrecent.deleteLocalFile()
				# Return Contents
				return manifest_contents

	return None


# Cleanup and Maintainance Functions

def finishSync(collection_name, wd_, backend_, db_session):
	'''Perform final actions for the backend. 1) Create a manifest file,
		2) Upload it to the server settings directory.'''
	
	manifest_ = sync.createFullSyncManifest(collection_name, wd_, db_session)
	backend_.uploadManifest(collection_name, 
		etree.tostring(manifest_, pretty_print=True))
	logging.info('Manifest Upload: ' + collection_name)
	logging.info('Finished Sync for ' + collection_name)