#!/usr/bin/python

import logging, time, base64
from lxml import etree
import os, sys, requests, posixpath
import urllib, urlparse

# Time Drive Imports
import sync
import sync.syncxml as syncxml
from sync.remote import RemoteResource

# Import Methods from DAV needed for ownCloud
from dav import createPropFind, createPropPatch, \
	sendMkCol, sendDeleteRequest, sendPropFind, \
	sendPropPatch, sendFileUpload, sendFileDownload, \
	davFolderContents, createWebFolder, checkWebFolder, \
	folderContents, deleteFile, downloadFile, \
	updateProperties, updateDavProperties, \
	downloadManifest, finishSync

# Import Data Properties
from dav import dav_prefixes

RT_Sync = True

# File Upload and Download Methods

def uploadFile(url_, user_, password_,
	contents_, mime_, dates_, localfile_url=None, file_md5=None):
	'''Upload a file to an owncloud WebDav server using a put request.
		Returns the file's remote etag if successful, otherwise returns None.'''

	logging.debug('Upload tile to ' + url_)

	# Ensure that there is data to send
	contents_size = len(contents_)
	if contents_size > 0:

		# Extract Creation and Modified Dates
		cdate_ = dates_[0]
		mdate_ = dates_[1]

		# Create DAV Request Headers
		request_headers = {'Content-Type' : mime_}
		if contents_:
			request_headers['Content-Length'] = str(len(contents_))
		elif localfile_url:
			request_headers['Content-Length'] = str(os.path.getsize(localfile_url))

		# Send Request
		r_, r_status = sendFileUpload(url_, user_, password_,
			request_headers, contents_, localfile_url)

		# Update the properties of the file
		prop_update = {'{'+dav_prefixes['D']+'}lastmodified' : str(int(sync.epochSeconds(mdate_)))}

		# Add md5 checksums to property list
		if file_md5:
			md5beg_, md5end_ = file_md5
			prop_update['{'+dav_prefixes['td']+'}md5beg'] = md5beg_
			prop_update['{'+dav_prefixes['td']+'}md5end'] = md5end_
		
		# Send Property Update
		updateDavProperties(url_, user_, password_, prop_update)

		# Check upload and retrieve the remote identifier for the file
		if r_status == True:
			rd_ = folderContents(url_, user_, password_)
			# Check to see if file object found
			if len(rd_) > 0: return rd_[0].file_etag

	return None


# Load and Parse the Remote File List

def addLocalDetailsToRdFolder(rdfolder_, wd_, topfolder_=False):
	'''Add local details to a remote file object that is a collection'''
	
	d_sep = sync.getDirectorySeparator()
	
	if topfolder_: 
		base_url, folder_name = os.path.split(wd_.strip(d_sep))
		rdfolder_.local_path = wd_
	else: 
		base_url, folder_name = posixpath.split(rdfolder_.remote_directory)
		rdfolder_.local_path = os.path.join(wd_, folder_name)
	
	rdfolder_.resource_name = folder_name
	
	return rdfolder_

def remoteFileList(collection_, backend_, wd_, rd_, 
		rdlist_ = None, include_folders=False):
	'''Retrieve the list of files and their propeties from the ownCloud
		website. Add them to a dictionary object indexed to the local path.'''

	# If no rdlist_, create a blank collection
	if rdlist_ is None:
		rdlist_ = {}

	# Get the contents for the current directory
	rd_contents = folderContents(rd_, 
		backend_.server_user, backend_.server_password)

	# Directory Separator for the System
	d_sep = sync.getDirectorySeparator()
	
	if rd_contents:
		
		# Iterate through contents, index to the local path
		for rdfile_ in rd_contents:
			
			# Add the local path to the remote resource

			# Copy localpath for folders (if the remote list should include folders)
			if (include_folders == True) & (rdfile_.resource_type == 'collection'):
				rdfile_ = addLocalDetailsToRdFolder(rdfile_, wd_)

			# Combine the localpath with the resource name for files
			else: rdfile_.local_path = os.path.join(wd_, rdfile_.resource_name)

			# Create a full url to compare against rd to make sure that a
			# reference to the rd is not mistakenly added to the list
			rdfile_fullpath = backend_.remoteFileUrl(rdfile_.full_path)

			# Check to see if the file is part of the collection 
			# and add to the remote file list
			if (rdfile_.local_path not in rdlist_.keys()) \
				& (rdfile_.local_path.rstrip(d_sep) != wd_) \
				& (rdfile_.local_path is not None) \
				& (rdfile_fullpath.rstrip('/') != rd_):
				rdlist_[rdfile_.local_path] = rdfile_
	
	return rdlist_

def remoteFileDetails(backend_, rd_):
	'''Retrieve a remote file object for a particular file. For retrieving the
		contents of a collection, use remoteFileList'''
	
	rd_contents = folderContents(rd_,
		backend_.server_user, backend_.server_password)
	if rd_contents: file_info = rd_contents[0]
	else: file_info = None
	return file_info

def remoteFolderChanges(collection_, backend_, rd_, wd_, db_session, 
		change_list = None, parent_dbfolder = None, parent_rdfolder = None):
	'''Analyze the contents of a remote folder and determine what files have changed'''

	# If no changelist supplied, create dictionary to keep track of modified files
	if change_list == None: changelist_ = []
	else: changelist_ = change_list
	folder_modified = 0

	# Check top-level directory for etag differences
	if (parent_dbfolder == None) & (parent_rdfolder == None):

		# Retrieve records from server and database
		top_rdfile = backend_.remoteFileDetails(rd_)
		top_rdfile = addLocalDetailsToRdFolder(top_rdfile, wd_, topfolder_=True)
		
		top_dbfile = sync.getLocalFolderRecord(wd_, db_session)

		# If eTag values match, return empty list of remote changes
		if (top_rdfile is not None) & (top_dbfile is not None):
			if top_rdfile.file_etag == top_dbfile.file_etag:
				logging.info('Top-Level folder etags match for collecion: ' \
					+ str(collection_['Collection-Name']))
				return changelist_

	# Compare files and subfolders for collections that differ

	# Retrieve list of files and folders from the server
	rd_filelist = backend_.remoteFileList(collection_, wd_, rd_, include_folders=True)

	# Iterate through files and folders, compare to etag in local database
	for resource_localname in rd_filelist.keys():

		# Retrieve resource from file list
		resource_ = rd_filelist.pop(resource_localname)

		# Compare folder etag values to those in the cache
		if resource_.resource_type == 'collection':
			analyze_folder = True
			dbfolder_ = sync.getLocalFolderRecord(resource_.local_path, db_session)
			
			if dbfolder_:
				if dbfolder_.file_etag == resource_.file_etag: analyze_folder = False

			# Analyze subfolders where remote and local etag values differ
			if analyze_folder:
				rd_sub = posixpath.join(rd_, resource_.resource_name)
				wd_sub = os.path.join(wd_, resource_.resource_name)
				logging.info('Looking for remote changes in ' + rd_sub)
				changelist_ = remoteFolderChanges(collection_, backend_, rd_sub, wd_sub, 
						db_session, change_list=changelist_, parent_rdfolder=resource_,
						parent_dbfolder=dbfolder_)

		# Compare files in directory to those in database
		else:
			dbfile_ = sync.getLocalFileRecord(resource_.local_path, db_session)

			if dbfile_:

				# Compare file etag values, add to changelist if modified
				if dbfile_.file_etag != resource_.file_etag:
					changelist_.append(resource_localname)
					folder_modified += 1

			# Unable to find database record, add to changelist
			else: 
				changelist_.append(resource_localname)
				folder_modified += 1

	# Update local folder records (parent folder)
	if (parent_rdfolder is not None) & (folder_modified == 0):
		logging.info('Updating Folder etag value: ' + parent_rdfolder.resource_name)
		sync.updateLocalFolderRecord_FromRemote(
			parent_rdfolder, parent_dbfolder, db_session)

	# Update local folder records (top level folder)
	if (parent_dbfolder == None) & (parent_rdfolder == None):
		if (top_rdfile is not None) & (len(changelist_) == 0):
			logging.info('Top-Level Folder Record: ' + top_rdfile.resource_name)
			sync.updateLocalFolderRecord_FromRemote(
				top_rdfile, top_dbfile, db_session)

	# Save any pending changes to the database
	db_session.commit()
	
	return changelist_