#!/usr/bin/python

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, DateTime, Boolean
from sqlalchemy.pool import StaticPool


import os, mimetypes, struct
from lxml import etree
from send2trash import send2trash

import sync

# Create Base (for data table classes)
Base = declarative_base()


## Database Methods ##

def createDataConnection(db_path):
	'''Create a connection to the database and opens a session. Returns
		both the underlying database engine and the session.'''

	# Create Database Engine
	dbconn_string = 'sqlite:///' + db_path
	engine = create_engine(dbconn_string,
		connect_args={'check_same_thread':False},
		poolclass=StaticPool, echo=False)

	# Create Session
	Session = sessionmaker(bind=engine)
	session = Session()

	return engine, session

def createDataTables(engine_):
	'''Check that the desired table exist'''
	Base.metadata.create_all(engine_)

def getLocalFileRecord(file_localpath, db_session):
	'''Query the local database and return the first non-deleted file'''
	dbfile_ = db_session.query(SyncFile).filter_by(
		full_path=file_localpath) \
		.filter((SyncFile.file_deleted != True) 
			| (SyncFile.file_deleted == None))\
		.first()
	return dbfile_

def getLocalFolderRecord(folder_localpath, db_session):
	'''Query the local database and return the first non-deleted folder record'''
	dbfolder_ = db_session.query(SyncFile).filter_by(
		full_path=folder_localpath) \
		.filter((SyncFile.file_deleted != True) 
			| (SyncFile.file_deleted == None))\
		.filter(SyncFile.directory == True) \
		.first()
	return dbfolder_


## Database Classes ##

class SyncFile(Base):
	'''Data related to local files to be included in the sync/backup'''
	__tablename__ = 'syncfiles'

	id = Column(Integer, primary_key=True)
	name = Column(String)
	directory = Column(Boolean)
	parent_directory = Column(String)
	full_path = Column(String)
	md5_beg = Column(String)
	md5_end = Column(String)
	file_etag = Column(String)
	cdate = Column(DateTime) # File cdate
	mdate = Column(DateTime) # File mdate
	file_deleted = Column(Boolean)
	ddate = Column(DateTime) # Date file removed from db

	def __init__(self, name, directory):
		'''Initialize the file. Requires the name of the file, 
			full path to its location on disk, and the date it was
			created (as a string).'''
		
		self.name = name
		self.parent_directory = directory
		self.full_path = os.path.join(self.parent_directory, self.name)

	def setMd5(self, md5_beg, md5_end):
		'''Sets the md5 signature for the beginning and end of the file.'''
		self.md5_beg = md5_beg
		self.md5_end = md5_end

	def setDateTimes(self, cdate_, mdate_):
		self.cdate = cdate_
		self.mdate = mdate_

	def setETag(self, file_etag):
		self.file_etag = file_etag

	def uploadFile(self, url_, sync_backend):
		'''Upload a file to a WebDav server using a put request'''

		# Retrieve the file mime type, encoding, and contents
		file_mime, file_enc = mimetypes.guess_type(self.name)
		file_ = open(self.full_path, 'rb')
		file_contents = file_.read()
		file_.close()

		# Determine Mimetype of File
		if file_mime is None:
			fname_, ext_ = os.path.splitext(url_)
			ext_ = ext_.lstrip('.')
			file_mime = 'application/' + ext_

		# Upload File
		r_ = sync_backend.uploadFile(url_, 
			file_contents, file_mime, 
			(self.cdate, self.mdate),
			file_md5 = (self.md5_beg, self.md5_end))
		if (r_ is not None) & (type(r_) == str):
			self.file_etag = r_
		
		return r_

	def deleteLocalFile(self):

		# Delete the local file
		if os.path.exists(self.full_path):
			send2trash(self.full_path)
			return True
		else: return False

	def xmlEntry(self, collection_):
		'''Create an Xml Entry that can be added to a sync manifest'''

		xml_root = etree.Element('SyncFile')
		# Item Path
		if self.full_path:
			xml_fullpath = etree.SubElement(xml_root, 'File-Path')
			xml_fullpath.text = sync.abs2rel(self.full_path, collection_)
		
		# Date Created/Date Modified
		if self.cdate:
			xml_cdate = etree.SubElement(xml_root, 'CDate')
			xml_cdate.text = sync.datetime2string(self.cdate)
		if self.mdate:
			xml_mdate = etree.SubElement(xml_root, 'MDate')
			xml_mdate.text = sync.datetime2string(self.mdate)
		
		# MD5 Signatures
		if self.md5_beg:
			xml_md5beg = etree.SubElement(xml_root, 'MD5-Beg')
			xml_md5beg.text = self.md5_beg
		if self.md5_end:
			xml_md5end = etree.SubElement(xml_root, 'MD5-End')
			xml_md5end.text = self.md5_end

		return xml_root

	def __repr__(self):
		'''How the data from the record should be represented at 
			the command line and in the debugger.'''
		return "<File('%s', '%s', '%s', '%s')>" % (self.name, 
			str(self.mdate), 
			self.md5_beg, self.md5_end)