#!/usr/bin/python

import posixpath, tempfile, os, shutil

from local import Base

from sqlalchemy import Column, Integer, String, DateTime, Boolean


## Helper Methods

def getRemoteCacheRecord(file_localpath, db_session):
	'''Query the local database and return the first non-deleted file'''
	rcfile_ = db_session.query(RemoteResource).filter_by(
		local_path=file_localpath)\
		.filter((RemoteResource.file_deleted != True) | 
			(RemoteResource.file_deleted == None))\
		.first()
	return rcfile_


## Remote Classes ##

class RemoteResource(Base):
	'''Data related to files stored on a remote server'''
	__tablename__ = 'remotefiles'

	# Name and File Paths
	id = Column(Integer, primary_key=True)
	resource_name = Column(String)
	remote_directory = Column(String)
	full_path = Column(String)
	local_path = Column(String)
	file_etag = Column(String)

	# Resouce Type and Status
	resource_type = Column(String)
	resource_status = Column(String)
	mdate = Column(DateTime)
	cdate = Column(DateTime)
	md5_beg = Column(String)
	md5_end = Column(String)
	file_deleted = Column(Boolean)
	ddate = Column(DateTime)

	# Path to Temp File with Contents
	resource_contents = Column(String)

	# Other Information
	file_operation = None

	def __init__(self, resource_name, rd_):
		'''Initialization function for the remote file. Requires a name and the remote
			directory where the file is stored.'''

		self.resource_name = resource_name
		self.remote_directory = rd_
		self.full_path = posixpath.join(rd_, resource_name)

	def setMd5(self, md5_beg, md5_end):
		'''Sets the md5 signature for the beginning and end of the file.'''
		self.md5_beg = md5_beg
		self.md5_end = md5_end

	def setEtag(self, etag_):
		'''Sets the file etag value'''
		self.file_etag = etag_

	def setDateTimes(self, cdate_, mdate_):
		self.cdate = cdate_
		self.mdate = mdate_

	def checkRemoteResource(self, sync_backend):
		'''Check to see if the remote resource exists'''
		return sync_backend.checkFile(self.full_path)

	def setResourceType(self, rtype_):
		self.resource_type = rtype_

	def setResourceStatus(self, status_):
		self.resource_status = status_

	def downloadFile(self, sync_backend):
		'''Download the file using the specified backend, save to a temporary file'''

		if self.full_path:
			file_contents = sync_backend.downloadFile(self.full_path)
			# Create a temporary file for the contents
			if file_contents:
				tfile_ = tempfile.NamedTemporaryFile(mode='w+b', delete=False)
				tfile_.file.write(file_contents)
				self.resource_contents = tfile_.name
				tfile_.close()
				return True
		
		return False

	def updateServerProperties(self, sync_backend):
		'''Update the file properties on the remote server'''

		property_list = {'md5beg' : self.md5_beg, 'md5end' : self.md5_end}
		return sync_backend.updateProperties(self.full_path, property_list)

	def deleteLocalFile(self):
		'''Delete the saved temporary file'''

		if os.path.exists(self.resource_contents):
			os.remove(self.resource_contents)
			self.file_contents = None
			return True
		else:
			return False

	def deleteRemoteFile(self, sync_backend):
		'''Delete the remote file'''

		result_ = sync_backend.deleteFile(self.full_path)
		if result_: return True
		else: return False

	def moveFile(self, dest_path):
		'''Move the file to the specified destination'''

		# Check to make sure that the destination directory exists
		if (os.path.exists(os.path.dirname(dest_path)) == True) \
			& (os.path.exists(self.resource_contents) == True):
			
			# Move file
			shutil.copyfile(self.resource_contents, dest_path)
			self.deleteLocalFile()
			return True
		
		else:
			logging.warning('Source or destination does not exist')
			return False