#!/usr/bin/python

import os, hashlib, base64, logging, datetime, posixpath
from local import SyncFile
from remote import RemoteResource
import sync

## Sync Methods ##

def downloadNewFile(rdfile_, db_session, sync_backend, server_file=False):
	'''Routine for downloading new files from the remote directory'''

	logging.info('Downloading new file from server: ' + rdfile_.resource_name)
	
	# Download file from server
	rd_downloadresult = rdfile_.downloadFile(sync_backend)
	rdfile_md5beg, rdfile_md5end = checkMd5(
		rdfile_.resource_contents, sync.FILE_CHUNK)

	# For files downloaded directly from the server, set the object md5
	# signature prior to verifying the download
	if server_file:
		rdfile_.setMd5(rdfile_md5beg, rdfile_md5end)
		# Update signatures on the server to prevent future sync problems
		rdfile_.updateServerProperties(sync_backend)

	# Md5 Matches Manifest, Move to Desttnation
	if (rdfile_md5beg == rdfile_.md5_beg) & (rdfile_md5end == rdfile_.md5_end):
		sync_backend.checkLocalFolder(
			os.path.dirname(rdfile_.local_path), create_folder = True)
		# Move file to destination
		rdfile_.moveFile(rdfile_.local_path)
		# Set adate and mdate
		os.utime(rdfile_.local_path, (
			sync.epochSeconds(datetime.datetime.utcnow()), # File Access Date, adate
			sync.epochSeconds(rdfile_.mdate))) # File Modification Date, mdate

		# Add the file to the local database
		resource_directory, resource_name = os.path.split(rdfile_.local_path)
		syncfile_ = SyncFile(resource_name, resource_directory)
		syncfile_.setDateTimes(rdfile_.cdate, rdfile_.mdate)
		syncfile_.setMd5(rdfile_md5beg, rdfile_md5end)
		syncfile_.setETag(rdfile_.file_etag)
		db_session.add(syncfile_)
		updateRemoteCache(rdfile_, db_session)
		db_session.commit()
		return True, None
	
	# Md5 different than manifest
	else:
		logging.warning('''MD5 signature for downloaded file doesn't match that in the manifest''')
		logging.warning('Web Resource - ' + rdfile_.resource_name + ':\n' \
			+ rdfile_.md5_beg + ', ' + rdfile_.md5_end)
		logging.warning('Download MD5:\n' + rdfile_md5beg + ', ' + rdfile_md5end)
		return False, None


def mergeExistingFile(file_localpath, file_url, file_dates, rdfile_, db_session, sync_backend):
	'''Routine used for merging a local file that exists into a remote collection.
		Requires the local filepath, the remote file url, a reference to the remote file
		metadata (remote file object), database session, and sync backend.'''

	# Retrieve the local file metadata
	md5_beg, md5_end = checkMd5(file_localpath, sync.FILE_CHUNK)

	# Compare the two files against each other
	if (md5_beg == rdfile_.md5_beg) & (md5_end == rdfile_.md5_end):

		# Add to database
		logging.info('Merging an existing file into the collection: ' + file_localpath)
		syncfile_folder, syncfile_name = os.path.split(file_localpath)
		syncfile_ = SyncFile(syncfile_name, syncfile_folder)
		syncfile_.setDateTimes(file_dates[0], file_dates[1])
		syncfile_.setMd5(md5_beg, md5_end)
		db_session.add(syncfile_)
		updateRemoteCache(rdfile_, db_session)
		db_session.commit()
		return True, None

	# If the local and remote files don't match, download the remote as a conflict,
	# allow the user to inspect and compare the two files
	else:
		logging.warning('Failed to merge an existing file into the collection: ' + file_localpath)
		logging.warning('The remote version of the file and the local file are different.')

		# If the local file is newer than the remote file
		if file_dates[1] > rdfile_.mdate:
			logging.warning('The remote version will be downloaded and added to the directory.')
			logging.warning('The local version will be uploaded to the server as canonical.')
			
			# Download the remote file, Create conflict filepath
			conflict_filepath = conflictFilepath(file_localpath, 
			datetime.datetime.now(), include_label='remote')
			rdfile_.downloadFile(sync_backend)
			rdfile_.moveFile(conflict_filepath)

			# Set the system access and modified times of the conflict file
			os.utime(conflict_filepath,
				(sync.epochSeconds(datetime.datetime.utcnow()),
				sync.epochSeconds(datetime.datetime.utcnow())))

			# Create a database entry and upload the local file to the server
			syncfile_folder, syncfile_name = os.path.split(file_localpath)
			syncfile_ = SyncFile(syncfile_name, syncfile_folder)
			syncfile_.setDateTimes(file_dates[0], file_dates[1])
			syncfile_.setMd5(md5_beg, md5_end)
			syncfile_.uploadFile(file_url, sync_backend)
			db_session.add(syncfile_)
			db_session.flush()

			# Update the Collection Cache
			rdfile_.md5_beg, rdfile_.md5_end = (md5_beg, md5_end)
			updateRemoteCache(rdfile_, db_session)
			db_session.commit()
			return False, conflict_filepath

		# Remote file newer than the local file
		elif file_dates[1] < rdfile_.mdate:
			logging.warning('The remote version will be downloaded and added to the directory.')
			logging.warning('The local version will be marked as a conflict.')

			# Download the remote file as temp
			rdfile_.downloadFile(sync_backend)

			# Check md5 checksum of downloaded file
			rdfile_md5beg, rdfile_md5end = checkMd5(
				rdfile_.resource_contents, sync.FILE_CHUNK)
			if (rdfile_md5beg == rdfile_.md5_beg) \
				& (rdfile_md5end == rdfile_.md5_end):

				# Create a conflict filepath, rename file to conflict
				conflict_filepath = conflictFilepath(file_localpath, 
					datetime.datetime.now(), include_label='local')
				os.rename(file_localpath, conflict_filepath)
				
				# Move downloaded file, set the local file properties
				rdfile_.moveFile(file_localpath)
				os.utime(file_localpath,
					(sync.epochSeconds(datetime.datetime.utcnow()),
					sync.epochSeconds(rdfile_.mdate)))

				# Add the downloaded file to the database
				syncfile_folder, syncfile_name = os.path.split(file_localpath)
				syncfile_ = SyncFile(syncfile_name, syncfile_folder)
				syncfile_.setDateTimes(datetime.datetime.utcnow(), rdfile_.mdate)
				syncfile_.setMd5(rdfile_md5beg, rdfile_md5end)
				db_session.add(syncfile_)
				updateRemoteCache(rdfile_, db_session)
				db_session.commit()
				return False, conflict_filepath

	return False, None


def uploadNewFile(file_localpath, file_url, file_dates, 
		file_md5, db_session, sync_backend):
	'''Routine used for uploading a new file from the local disk to the server
		and logging it in the database.'''
	
	logging.info('Uploading a new file to the server: ' + file_localpath)

	# Create a new SyncFile record
	syncfile_folder, syncfile_name = os.path.split(file_localpath)
	syncfile_ = SyncFile(syncfile_name, syncfile_folder)
	syncfile_.setDateTimes(file_dates[0], file_dates[1])
	syncfile_.setMd5(file_md5[0], file_md5[1])
 
	# Upload the file to the server
	r_ = syncfile_.uploadFile(file_url, sync_backend)

	# Upload Successful
	if r_:

		# Add to the database and save
		db_session.add(syncfile_)
		db_session.commit()
		return True, None

	# Upload Unsuccessful
	return False, None


def uploadUpdatedFile(dbfile_, file_url, file_dates, file_md5, 
		db_session, sync_backend, remote_missing=False):
	'''Routine used for uploading an updated version of the file to the remote server'''

	# Upload file to server
	try:

		# Submit changes to the database - Local File
		dbfile_.setDateTimes(file_dates[0], file_dates[1])
		dbfile_.setMd5(file_md5[0], file_md5[1])

		r_ = dbfile_.uploadFile(file_url, sync_backend)
		logging.info('Updated file uploaded to server: ' + dbfile_.name)

		# Submit changes to the database - Remote Cache
		rd_filefolder, rd_filename = posixpath.split(file_url)
		rdfile_ = RemoteResource(rd_filename, rd_filefolder)
		rdfile_.local_path = dbfile_.full_path
		rdfile_.setDateTimes(file_dates[0], file_dates[1])
		rdfile_.setMd5(file_md5[0], file_md5[1])
		updateRemoteCache(rdfile_, db_session)

		db_session.commit()
		return True, None

	except: return False, None


def replaceLocalFile(dbfile_, rdfile_, db_session, sync_backend):
	'''Routine used for downloading a file from the remote server and replacing
		the local file. Checks the mdate of all three files (file on disk, file
		in the database, and the file on the server), checks the checksums of the
		files to make sure that they all match. If there is a problem, it creates
		a conflict file and marks it in the log. Requires the local file path,
		a dbfile_ object, an rdfile_ object, a link to the db_session, and a link
		to the sync_backend.'''

	logging.info('Updated file on server: ' + rdfile_.resource_name)

	# Make sure local checksum match before downloading file
	md5_beg, md5_end = checkMd5(dbfile_.full_path, sync.FILE_CHUNK)
	if (md5_beg == dbfile_.md5_beg) & (md5_end == dbfile_.md5_end):

		# Download file as temp
		rdfile_.downloadFile(sync_backend)
		
		# Check md5 checksum of downloaded file
		rdfile_md5beg, rdfile_md5end = checkMd5(
			rdfile_.resource_contents, sync.FILE_CHUNK)
		if (rdfile_md5beg == rdfile_.md5_beg) \
			& (rdfile_md5end == rdfile_.md5_end):
			
			# Move file to destination
			rdfile_.moveFile(dbfile_.full_path)
			# Set adate and mdate
			os.utime(dbfile_.full_path, (
				sync.epochSeconds(datetime.datetime.utcnow()),
				sync.epochSeconds(rdfile_.mdate)))
			
			# Update the database record with date, md5 hash
			dbfile_.mdate = rdfile_.mdate
			dbfile_.setMd5(rdfile_md5beg, rdfile_md5end)
			updateRemoteCache(rdfile_, db_session)
			db_session.commit()
			
			return True, None

		# If md5 checksum of download different than that 
		# in manifest, create a new file and mark as a conflict
		# Upload local file as canonical
		else:

			logging.warning('File mismatch, the file on the server is different than that in the manifest. Uploading the file from the local computer as canonical.')
			
			# Create a new conflict file in the same directory, 
			# use the timestamp from remote
			conflict_filepath = conflictFilepath(dbfile_.full_path, rdfile_.mdate)

			# Move the file to the new file location
			rdfile_.moveFile(conflict_filepath)
			# Set adate and mdate
			os.utime(conflict_filepath, (
				sync.epochSeconds(datetime.datetime.utcnow()),
				sync.epochSeconds(rdfile_.mdate)))

			# Upload the file on disk as canonical
			dbfile_.uploadFile(rdfile_.full_path, sync_backend)
			return False, conflict_filepath

	# If local checksums don't match, contine with download but mark the local file
	# as a conflict so that the user can take a look at it
	else:

		logging.warning('Local file checksums do not match. Marking the local copy as a conflict and downloading the remote file.')

		# Create a new conflict file in the same directory
		# use the timestamp from the local database
		conflict_filepath = conflictFilepath(dbfile_.full_path, dbfile_.mdate)

		# Rename the file on the local disk
		os.rename(dbfile_.full_path, conflict_filepath)

		# Download the file from the server, move to the local file path
		rdfile_.downloadFile(sync_backend)
		rdfile_.moveFile(dbfile_.full_path)

		# Update information in the database record
		rmd5_beg, rmd5_end = checkMd5(dbfile_.full_path, sync.FILE_CHUNK)
		# Update the utimes for the file on disk
		os.utime(dbfile_.full_path, (
				sync.epochSeconds(datetime.datetime.utcnow()),
				sync.epochSeconds(rdfile_.mdate)))
		dbfile_.mdate = rdfile_.mdate
		dbfile_.setMd5(rmd5_beg, rmd5_end)
		db_session.commit()

		return False, conflict_filepath


def deleteRemoteFile(rdfile_, dbfile_, db_session, sync_backend):
	'''Delete the remote file, mark the file as deleted in the database.'''
	
	if (rdfile_ is not None):
		
		# Add comment to the log
		logging.info('Deleting remote file: ' + rdfile_.full_path+'\n')
		
		# Delete the remote file
		result_ = rdfile_.deleteRemoteFile(sync_backend)
	
	else: result_ = False

	if result_: 
		
		# Mark as deleted in the database, set deletion date/time
		dbfile_.file_deleted = True
		dbfile_.ddate = datetime.datetime.utcnow()
		updateRemoteCache(rdfile_, db_session, mark_deleted=True)
		db_session.commit()
	
	return result_, None


def deleteLocalFile(dbfile_, rdfile_, db_session):
	'''Delete the local copy of the file, mark the file as deleted in database.'''
	
	logging.info('Deleting local file: ' + dbfile_.full_path)
		
	# Delete the local file
	result_ = dbfile_.deleteLocalFile()
	if result_:

		# Mark as deleted in the database, set the deletion date/time
		dbfile_.file_deleted = True
		dbfile_.ddate = datetime.datetime.utcnow()
		updateRemoteCache(rdfile_, db_session, mark_deleted=True)
		db_session.commit()
		return True, None

	else: return False, None


def moveRemoteFile(rdfile_, destination_, sync_backend):
	'''Move the remote file to a new location'''
	pass


def missingManifestFile(dbfile_, server_url, sync_backend, db_session):
	'''Found a local file that is not in the manifest. 
		Check for remote, see if the local file should be deleted.'''

	logging.info('Found local file without entry in the manifest: ' + dbfile_.name)
	logging.info('Checking remote file')

	# Create remote resource and check to see if it exists
	file_localdir, file_localname = os.path.split(dbfile_.full_path)
	rdfile_ = RemoteResource(file_localname, server_url)

	# Check to see if the remote file exists
	if rdfile_.checkRemoteResource(sync_backend):

		# Download the file and calculate md5
		rdfile_.downloadFile(sync_backend)
		remote_md5_beg, remote_md5_end = checkMd5(
			rdfile_.resource_contents, sync.FILE_CHUNK)
		
		# Remote and local versions of the file are the same
		if (remote_md5_beg == dbfile_.md5_beg) \
			& (remote_md5_end == dbfile_.md5_end):
			logging.info('Remote version is the same as local version.')
			logging.info('It will be added to the manifest at the end of the session.')
			# Update the remote cache
			rdfile_.setMd5(remote_md5_beg, remote_md5_end)
			updateRemoteCache(rdfile_, db_session)
			db_session.commit()
			return True, None
		
		# Remote and local versions are different
		# Mark the remote version as a conflict
		# Upload the local version as canonical
		else:
			logging.info('Remote version is different than local version')
			logging.info('The remote version has been downloaded and marked as a conflict.')
			logging.info('The local version will be uploaded as the canonical version of the file.')
			
			# Create a conflict file for the remote version
			conflict_filepath = conflictFilepath(file_localpath, datetime.datetime.now())
			rdfile_.moveFile(conflict_filepath)

			# Set the system access and modified times
			os.utime(conflict_filepath,
				sync.epochSeconds(datetime.datetime.utcnow()),
				sync.epochSeconds(datetime.datetime.utcnow()))

			# Upload the local version as canonical
			dbfile_.uploadFile(rdfile_.full_path, sync_backend)

			# Update the remote cache
			rdfile_.setDateTimes(dbfile_.cdate, dbfile_.mdate)
			rdfile_.setMd5(dbfile_.md5_beg, dbfile_.md5_end)
			updateRemoteCache(rdfile_, db_session)
			db_session.commit()

			return False, conflict_filepath


def downloadRemoteAsConflict(rdfile_, file_localpath, sync_backend):
	'''Download a remote file as a conflict file to allow the user an opportunity 
		to resolve problems found during the file synchronization'''

	# Create a new conflict file in the same directory as the local file
	# use the timestamp from the remote file
	conflict_filepath = conflictFilepath(file_localpath, rdfile_.mdate)
	rdfile_.downloadFile(sync_backend)
	rdfile_.moveFile(conflict_filepath)

	# Update the utimes for the file on disk
	os.utime(file_localpath, (
		sync.epochSeconds(datetime.datetime.utcnow()), 
		sync.epochSeconds(rdfile_.mdate)))
	return True, conflict_filepath


def conflictFilepath(file_localpath, datetime_, include_label=None):
	'''Create a new filepath to be used for conflict files from the original file
		and a datetime object.'''
	conflict_directoryname, conflict_originalname = os.path.split(file_localpath)
	conflict_filebase, conflict_fileext = os.path.splitext(conflict_originalname)
	# Include timestamp and remote name
	if include_label: conflict_filename = conflict_filebase + '-conflict.' \
		+ datetime_.strftime('%Y-%m%d.%H%M') + '-' + str(include_label) + conflict_fileext 
	# Include timestamp
	else: conflict_filename = conflict_filebase + 'conflict.' \
		+ datetime_.strftime('%Y-%m%d.%H%M') + conflict_fileext
	return os.path.join(conflict_directoryname, conflict_filename)


def checkMd5(file_path, chunk_len):
	'''Calculate an MD5 checksum for the beginning and end of a file.
		The checksums are used to determine if the file has been modified.'''

	# Check to ensure that the file exists
	if not os.path.exists(file_path):
		return None, None

	# Retrieve a chunk at the beginning and end of the file
	f_ = open(file_path, 'rb')
	fchunk_beg = f_.read(chunk_len)

	# Retrieve a chunk of text the same size from the end of the file
	fchunk_end = bytes('')
	for line in reversed(f_.readlines()):
		fchunk_end += bytes(line)
		if len(fchunk_end) > chunk_len:
			fchunk_end = fchunk_end[:chunk_len]
			break

	# Create hash of the beginning and end of the file
	hash_beg = hashlib.md5(fchunk_beg).digest()
	hash_end = hashlib.md5(fchunk_end).digest()
	return base64.b64encode(hash_beg), base64.b64encode(hash_end)


def updateLocalFolderRecord_FromLocal(folder_localpath, folder_dates,
		dbfolder_, db_session, mark_deleted=False):
	'''Add local folder information to the local database, or update the 
		record of an existing folder.'''

	# For folders not yet added to the database
	if dbfolder_ == None:
		folder_parentdir, folder_name = os.path.split(
			folder_localpath.rstrip(sync.getDirectorySeparator()))
		dbfolder_ = SyncFile(folder_name, folder_parentdir)
		dbfolder_.directory=True
		dbfolder_.setDateTimes(*folder_dates)
		db_session.add(dbfolder_)
		db_session.flush()

	else:
		# Update mdate and cdate of folder
		if not (folder_dates[0] == dbfolder_.cdate) \
			| (folder_dates[1] == dbfolder_.mdate):
			dbfolder_.setDateTimes(*folder_dates)


def updateLocalFolderRecord_FromRemote(rdfolder_, dbfolder_, 
		db_session, mark_deleted=False):
	'''Add remote folder information to the local database, or update the 
		record of an existing folder.'''

	# For folders not yet added to the database
	if dbfolder_ == None:
		dbfolder_ = SyncFile(rdfolder_.resource_name, os.path.dirname(rdfolder_.local_path))
		dbfolder_.directory =True
		dbfolder_.file_etag = rdfolder_.file_etag
		db_session.add(dbfolder_)
		db_session.flush()

	else:
		if dbfolder_.file_etag is not rdfolder_.file_etag:
			dbfolder_.file_etag = rdfolder_.file_etag


def updateRemoteCache(rdfile_, db_session, mark_deleted = False):
	'''Add remote file information to the cache of remote files, or
		update the file's existing information.'''

	# Try to retrieve the record from the database
	cfile_ = db_session.query(RemoteResource).filter_by(
			local_path=rdfile_.local_path)\
		.filter((RemoteResource.file_deleted != True) | 
			(RemoteResource.file_deleted == None))\
		.first()

	# Add Remote Resource to Database
	if cfile_ is None:
		db_session.add(rdfile_)
		db_session.flush()

	# Compare the local and remote resources, update differences
	else:
		if rdfile_.cdate is not cfile_.cdate:
			cfile_.cdate = rdfile_.cdate
		if rdfile_.mdate is not cfile_.mdate:
			cfile_.mdate = rdfile_.mdate
		if rdfile_.md5_beg is not cfile_.md5_beg:
			cfile_.md5_beg = rdfile_.md5_beg
		if rdfile_.md5_end is not cfile_.md5_end:
			cfile_.md5_end = rdfile_.md5_end
		if rdfile_.file_etag is not cfile_.file_etag:
			cfile_.file_etag = rdfile_.file_etag

	if mark_deleted:
		if cfile_ is None:
			rdfile_.file_deleted = True
			rdfile_.ddate = datetime.datetime.utcnow()
		else:
			cfile_.file_deleted = True
			cfile_.ddate = datetime.datetime.utcnow()