#!/usr/bin/python

import os, sys, logging
import datetime, re, posixpath
from lxml import etree
import sync

# Time-Drive Imports
from remote import RemoteResource

## Helper Functions ##

def removeXmlDeclarations(xml_string):
	'''Helper function which looks for and removes <?xml> delcarations'''
	xml_dec = re.compile(r'encoding=".*"')
	return xml_dec.sub('', xml_string)

def trimNamespace(xml_string):
	'''Helper function which removes namespaces from etree tags'''
	re_ns = re.compile(r'\{.*\}')
	return re_ns.sub('', xml_string)

def findHtmlStatus(xml_string):
	'''Helper function which locates the HTML status from a WebDav request'''
	re_status = re.compile(r'\d{3}')
	status_ = re_status.findall(xml_string)
	if len(status_) > 0:
		return status_[0]
	else:
		return None


## Manifest Functions ##

def parseManifest(manifest_text, wd_, rd_):
	'''Parse an XML manifest and returns a dictionary of file-paths
		and linked RemoteResource objects for their equivalent files
		on the server. Requires the manifest text, the full path to the
		remote directory and the local working directory, all as strings.'''

	collection_files = {}

	# Try and parse the XML file
	try:
		xml_ = etree.XML(removeXmlDeclarations(manifest_text))

	# Upon error, notify the user and return an empty dictionary
	except:
		logging.warning('Unable to parse XML file')
		logging.warning(manifest_text)
		return collection_files

	# Find the collection properties
	xml_properties = xml_.find('Properties')
	xml_contents = xml_.find('Contents')

	for file_ in xml_contents:
		file_remote = parseManifestFileXml(file_, wd_, rd_)
		if file_remote:
			collection_files[file_remote.local_path] = file_remote

	return collection_files

def parseManifestFileXml(file_xml, wd_, rd_):
	'''Creates a RemoteResource object from a block of XML corresponding
		to a single manifest file. Also requires the local working directory,
		and the full path to the remote directory where the collection is
		stored, both as strings.'''

	# Placeholder variables for the file details
	file_name = None
	file_rd = None
	file_fullpath = None
	file_localpath = None
	file_mdate = None
	file_cdate = None
	file_md5beg = None
	file_md5end = None

	xml_filename = file_xml.find('File-Path')
	xml_cdate = file_xml.find('CDate')
	xml_mdate = file_xml.find('MDate')
	xml_md5beg = file_xml.find('MD5-Beg')
	xml_md5end = file_xml.find('MD5-End')

	if xml_filename is not None:

		# File Name, Remote and Local Paths
		rdfolder_, file_name = posixpath.split(xml_filename.text)
		if rdfolder_:
			rdfolder_fullpath = posixpath.join(rd_, rdfolder_)
			file_localpath = os.path.abspath(os.path.join(wd_, rdfolder_, file_name))
		else:
			rdfolder_fullpath = rd_
			file_localpath = os.path.join(wd_, file_name)

		# Create a RemoteResource Object
		file_remoteresource = RemoteResource(file_name, rdfolder_fullpath)
		file_remoteresource.local_path = file_localpath

		# Determine File Dates and Times
		if xml_cdate is not None:
			file_cdate = sync.string2datetime(xml_cdate.text)
		if xml_mdate is not None:
			file_mdate = sync.string2datetime(xml_mdate.text)
		file_remoteresource.setDateTimes(file_cdate, file_mdate)

		# MD5 Checksums
		if xml_md5beg is not None:
			file_md5beg = xml_md5beg.text
		if xml_md5end is not None:
			file_md5end = xml_md5end.text
		file_remoteresource.setMd5(file_md5beg, file_md5end)

		return file_remoteresource

	else:
		return None

## Xml Request/Response Skeletons ##

def syncManifest(collection_, datetime_):
	'''Create the etree outline of a sync response'''

	sync_root = etree.Element('Sync')
	sync_prop = etree.SubElement(sync_root, 'Properties')
	sync_contents = etree.SubElement(sync_root, 'Contents')

	# Collection Data
	sync_program = etree.SubElement(sync_prop, 'Program')
	sync_program.text = 'Time-Drive'
	sync_collection = etree.SubElement(sync_prop, 'Collection')
	sync_collection.text = collection_
	sync_datetime = etree.SubElement(sync_prop, 'SyncTime')
	sync_datetime.text = sync.datetime2string(datetime_)

	return sync_root, sync_prop, sync_contents