import sys, os, logging, posixpath, datetime
from collections import deque
from PyQt4 import QtCore, QtGui

import timedrive
from timedrive.background import FolderWatcher, SyncTask

import sync
from sync import local, remote, syncaction, backend

class TimeDriveSystemTray(QtGui.QWidget):
	'''Background class used for updating the backup collection 
		in real time. Runs as a QApplication daemon with a 
		watchdog process.'''

	def __init__(self):
		'''Initialize the background class, setup up directory
			watchers for the directories in the sync 
			collections file.'''
		
		super(TimeDriveSystemTray, self).__init__()
		self.threadpool = {}
		self.sync_db = {}

		# Active OPS List and Transfer Queues
		self.active_ops = {}
		self.checks = deque([])
		self.remote = deque([])
		self.uploads = deque([])
		self.downloads = deque([])

		# Load Sync Settings and Collections
		self.syncsettings = sync.appSettings('resources')
		self.sync_collections = sync.collectionSettings('resources')

		# Create System Tray Actions and Icon
		self.createActions()
		self.setupUi()

		# Set Up Logging for the Application
		log = logging.getLogger()
		log.setLevel(logging.INFO)
		
		# Create Background Watcher
		self.watcher = FolderWatcher()
		self.watcher.watchpaths = {}
		
		# Collect Local Filepaths to Watch
		for collection_ in self.sync_collections:
			collection_localpath = self.sync_collections[collection_]['Local-Path']
			self.watcher.watchpaths[collection_] = collection_localpath
		
		# Connect Signals
		self.connect(QtGui.QApplication.instance(),
			QtCore.SIGNAL('aboutToQuit()'), self.cleanUpSync)
		self.connect(self.watcher,
			QtCore.SIGNAL('resourceMoved(bool, QString, QString)'),
			self.resourceMoved)
		self.connect(self.watcher,
			QtCore.SIGNAL('resourceCreated(bool, QString)'),
			self.resourceCreated)
		self.connect(self.watcher,
			QtCore.SIGNAL('resourceDeleted(bool, QString)'),
			self.resourceDeleted)
		self.connect(self.watcher,
			QtCore.SIGNAL('resourceModified(bool, QString)'),
			self.resourceModified)
		self.watcher.start()

		self.setWindowTitle("Time Drive")
		self.resize(400, 300)

		# Setup Sync Backend
		server_ = self.syncsettings['AppSettings']['Scheme'] + '://' \
			+ self.syncsettings['AppSettings']['Host']
		server_url = posixpath.join(server_, self.syncsettings['AppSettings']['Path'])
		self.sync_backend = backend.SyncBackend(server_url, #Url of Server
			self.syncsettings['AppSettings']['Username'], # Username
			self.syncsettings['AppSettings']['Password'], # Password
			self.syncsettings['AppSettings']['Backend']) # Sync Backend to Use
		self.sync_backend.setExcludeList(
			self.syncsettings['AppSettings']['Exclude'])
		self.sync_backend.settingsFolder(
			self.syncsettings['AppSettings']['Remote-Settings'])

		# Setup connections to sync databases
		for collection_ in self.sync_collections:
			db_path = os.path.join(self.syncsettings['AppSettings']['Local-Settings'],
				self.sync_collections[collection_]['Database'])
			db_engine, db_session = sync.createDataConnection(db_path)
			# Ensure database tables exist
			sync.createDataTables(db_engine)
			# Store the reference to the session and engine for later use
			self.sync_db[collection_] = [db_engine, db_session]
			self.localCollectionChanges(collection_)

		# Create timer to check periodically for updates
		self.createActionTimer(self.syncsettings['AppSettings']['LocalCheckInterval'],
			self.checkAllCollections, single_shot=False)

		# Start Local Folder Checks
		self.startCheck()

	def checkAllCollections(self):
		'''Add entries for all collections in the settings list'''
		for collection_ in self.sync_collections:
			self.localCollectionChanges(collection_)
		self.startCheck()

	def localCollectionChanges(self, collection_):
		'''Add a folder check entry in the checks queue for all collections'''
		logging.info('Adding collection check to the queue: ' + collection_)
		if self.sync_collections[collection_]['Local-Path'] \
			not in self.active_ops.keys():
			self.checks.append(
				(str(collection_), sync.LOCAL_CHECK_FOLDER_CHANGES,
					self.sync_collections[collection_]['Local-Path']))
			self.active_ops[
				self.sync_collections[collection_]['Local-Path']] = sync.CHECK

	def closeEvent(self, event):
		'''Upon closing the window, provide the user with a warning'''
		if self.trayIcon.isVisible():
			QtGui.QMessageBox.information(self, "System Tray",
				"The program will keep running in the system tray."
				"To terminate the program, choose 'Quiz'"
				"in the context menu of the system tray.")
			self.hide()
			event.ignore()

	def createActions(self):
		'''Create the status bar action icons.'''
		self.quitAction = QtGui.QAction("Quit", 
			self, triggered = QtGui.qApp.quit)

	def createActionTimer(self, duration_, next_action_event, single_shot=True):
		'''Create a timer for an action'''
		
		action_timer = QtCore.QTimer(self)
		action_timer.setSingleShot(single_shot)
		self.connect(action_timer,
			QtCore.SIGNAL('timeout()'),
			next_action_event)
		action_timer.start(
			int(duration_))

	def setupUi(self):
		'''Create the system tray menu, add actions, add icon
			to the system tray.'''

		# Create separate process for Windows, set window icon
		if sys.platform == 'win32':
			import ctypes
			ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(
				str(self.syncsettings['AppSettings']['AppName']))
		self.setWindowIcon(QtGui.QIcon('resources/time-drive.png'))

		# Create system tray task menu, Add actions
		self.trayIconMenu = QtGui.QMenu(self)
		self.trayIconMenu.addSeparator()
		self.trayIconMenu.addAction(self.quitAction)

		# Create Icon, Add to the System tray
		self.trayIcon = QtGui.QSystemTrayIcon(self)
		self.trayIcon.setContextMenu(self.trayIconMenu)
		if sys.platform == 'win32':
			t_icon = QtGui.QIcon('resources/time-drive.png')
		else: t_icon = QtGui.QIcon('resources/task-tray.svg')
		self.trayIcon.setIcon(t_icon)
		self.trayIcon.setVisible(True)

	def showTrayMessage(self, message_, detail_):
		icon = QtGui.QSystemTrayIcon.MessageIcon(1)
		self.trayIcon.showMessage(message_,
			detail_, icon, 1000)

	def processTaskResults(self, taskname_):
		'''Process task results and execute any other needed tasks to finish
			the desired sync operation.'''

		# Remove the task from the threadpool
		sync_task = self.threadpool.pop(taskname_)
		sync_task.task_name + ' has completed'

		# Retrieve task results
		if sync_task.results:
			collection_, next_task_type, file_localpath = sync_task.results
		else: collection_, next_task_type, file_localpath = (None, None, None)

		# Local and Remote File Checks
		if str(taskname_) in [sync.LOCAL_CHECK_NEW, sync.LOCAL_CHECK_EXISTING,
			sync.LOCAL_CHECK_DELETION, sync.LOCAL_CHECK_FOLDER_CHANGES]:
			next_action = True

			# File Uploads
			if (next_task_type in [sync.RT_UPLOAD_NEW, sync.RT_UPLOAD_UPDATE]):
				self.uploads.append((collection_, next_task_type, file_localpath))
				self.startFileUpload()

			# Remote Operations
			elif (next_task_type in [sync.RT_DELETE_REMOTE, sync.RT_MOVE_REMOTE]):
				self.remote.append((collection_, next_task_type, file_localpath))
				self.startRemoteCollectionOp()

			# Successful Operations
			elif (next_task_type in [sync.RT_SUCCESS, sync.RT_NOTNEEDED]):
				
				# Process the results from a local folder difference check
				if str(taskname_) in [sync.LOCAL_CHECK_FOLDER_CHANGES]:

					# Add files to upload list for processing
					for file_ in sync_task.file_list:
						self.checks.append((collection_, sync.LOCAL_CHECK_EXISTING, file_))
						self.active_ops[file_] = sync.UPDATE
					
				self.active_ops.pop(file_localpath)

			elif (next_task_type in [sync.RT_FAIL]):
				self.checks.append((
					collection_, str(taskname_), file_localpath))
				self.createActionTimer(sync.syncsettings['AppSettings']['FailedLocalCheckPause'],
					self.startCheck)


			if next_action: self.startCheck()

		# File Uploads and Remote Collection Operations
		elif (str(taskname_) in [sync.RT_UPLOAD_NEW, sync.RT_UPLOAD_UPDATE,
			sync.RT_DELETE_REMOTE, sync.RT_MOVE_REMOTE]):

			# Determine what type of action needs to be started (upload, download,
			#	or remote collection operation)
			if str(taskname_) in [sync.RT_UPLOAD_NEW, sync.RT_UPLOAD_UPDATE]:
				next_task_method = self.startFileUpload
			elif str(taskname_) in [sync.RT_DELETE_REMOTE, sync.RT_MOVE_REMOTE]:
				next_task_method = self.startRemoteCollectionOp

			# Process Results: Successful, Not Needed, or Failed

			# Action Successful or Not Needed, Remove from Active Queue
			if next_task_type in [sync.RT_SUCCESS, sync.RT_NOTNEEDED]:
				self.active_ops.pop(file_localpath)
				next_task_method()
			
			# Action Unsuccessful
			else:

				# Add failed event to the appropriate queue to run again
				if str(taskname_) in [sync.RT_UPLOAD_NEW, sync.RT_UPLOAD_UPDATE]:
					self.uploads.append((
						collection_, str(taskname_), file_localpath))
				elif str(taskname_) in [sync.RT_DELETE_REMOTE, sync.RT_MOVE_REMOTE]:
					self.remote.append((
						collection_, str(taskname_), file_localpath))
				
				# Add note to the log of the failed operation
				logging.warning('Unable to '+str(taskname_)+' for ' \
					+ str(file_localpath)\
					+ '. Appending to queue.')
				
				# Create timer to try action again
				logging.warning('Retry failed operation: ' + str(taskname_))
				self.createActionTimer(
					self.syncsettings['AppSettings']['FailedUploadPause'], next_task_method)

		print self.active_ops
		print self.threadpool
		print self.remote


	def startRemoteCollectionOp(self):
		'''Begin an operation which affects the remote collection: remote file
			rename, remote file deletion, remote folder sync.'''
		
		remoteop_running = False
		for remote_optype in [sync.RT_DELETE_REMOTE, sync.RT_MOVE_REMOTE]:
			if remote_optype in self.threadpool.keys():
				remoteop_running = True

		if (remoteop_running is False) & (len(self.remote) > 0):

			# Retrieve the collection, type of upload, and path of the file
			collection_, sync_taskname, file_localpath = self.remote.popleft()

			# Retrieve the db_engine and db_session for the collection
			try: db_engine, db_session = self.sync_db[collection_]
			except: db_engine, db_session = (None, None)

			# Determine which method needs to be executed
			if sync_taskname == sync.RT_DELETE_REMOTE:
				sync_task_method = timedrive.realtime.deleteRemoteFile
				sync_task_args = [file_localpath, db_engine, db_session,
					self.sync_collections[collection_], self.sync_backend]
			
			sync_task = SyncTask(sync_taskname, 
				sync_task_method, sync_task_args)
			self.startBackgroundOp(sync_task)

	def startFileUpload(self):
		'''Begin a file upload background operation'''

		# Check to see if there is an upload already running 
		fileupload_running = False
		for upload_type in [sync.RT_UPLOAD_NEW, sync.RT_UPLOAD_UPDATE]:
			if upload_type in self.threadpool.keys():
				fileupload_running = True

		# If no upload is running, start one from the queue
		if (fileupload_running is False) & (len(self.uploads) > 0):

			# Retrieve the collection, type of upload, and path of the file
			collection_, sync_taskname, file_localpath = self.uploads.popleft()

			# Retrieve the db_engine and db_session for the collection
			try: db_engine, db_session = self.sync_db[collection_]
			except: db_engine, db_session = (None, None)

			if sync_taskname == sync.RT_UPLOAD_NEW: sync_newfile = True
			else: sync_newfile = False

			# Create Task and Add to Threadpool
			sync_task = SyncTask(sync_taskname,
				timedrive.realtime.uploadFile,
				[file_localpath, db_engine, db_session,
					self.sync_collections[collection_],
					self.sync_backend, sync_newfile])
			self.startBackgroundOp(sync_task)

	def startCheck(self):
		'''Check to see if there is an active check running,
			if not start a new resource check'''

		# Look at existing threadpool to see if there is a check running
		check_running = False
		for checktype_ in [sync.LOCAL_CHECK_NEW, sync.LOCAL_CHECK_EXISTING,
			sync.LOCAL_CHECK_DELETION]:
			if checktype_ in self.threadpool.keys():
				check_running = True

		# If no check is running, start one from the queue
		if (check_running is False) & (len(self.checks) > 0):
			sync_task = None
			
			# Retrieve the collection, type of check, and path of the file
			collection_, sync_taskname, file_localpath = self.checks.popleft()
			
			# Retrieve the db_engine and db_session for the collection
			try: db_engine, db_session = self.sync_db[collection_]
			except: db_engine, db_session = (None, None)

			# New Files
			if sync_taskname == sync.LOCAL_CHECK_NEW:
				sync_task = SyncTask(sync.LOCAL_CHECK_NEW, 
					timedrive.realtime.localFileMod,
					[file_localpath, db_engine, db_session,
						self.sync_collections[collection_], True],
					parent=self)
			
			# Existing Files
			elif sync_taskname == sync.LOCAL_CHECK_EXISTING:
				print self.sync_collections
				sync_task = SyncTask(sync.LOCAL_CHECK_EXISTING, 
					timedrive.realtime.localFileMod,
					[file_localpath, db_engine, db_session,
						self.sync_collections[collection_]],
					parent=self)

			# File Deletions
			elif sync_taskname == sync.LOCAL_CHECK_DELETION:
				sync_task = SyncTask(sync.LOCAL_CHECK_DELETION,
					timedrive.realtime.checkLocalFileDelete,
					[file_localpath, db_engine, db_session,
						self.sync_collections[collection_]],
					parent=self)

			# Check Folder for Changes
			elif sync_taskname == sync.LOCAL_CHECK_FOLDER_CHANGES:
				sync_task = SyncTask(sync.LOCAL_CHECK_FOLDER_CHANGES,
					timedrive.realtime.localFolderChanges,
					[self.sync_collections[collection_]['Local-Path'],
						db_engine, db_session, 
						self.sync_collections[collection_]],
					parent=self)

			# Add task to the thread pool
			if sync_task: self.startBackgroundOp(sync_task)

	def startBackgroundOp(self, sync_task):
		'''Add a background operation to the threadpool and connect events'''

		self.threadpool[QtCore.QString(sync_task.task_name)] = sync_task
		self.connect(sync_task, 
			QtCore.SIGNAL('taskComplete(QString)'), self.processTaskResults)
		sync_task.start()

	# File Operation Watch Events

	def resourceMoved(self, folder_, src_path, dest_path):
		'''Resource moved from one location to another.'''
		
		# Retrieve collection and reference to the database
		collection_ = timedrive.findCollection(
			self.sync_collections, dest_path)
		try: db_engine, db_session = self.sync_db[collection_]
		except: db_engine, db_session = (None, None)

		# Add description of event to the log
		resource_type = 'folder' if folder_ else 'file'
		logging.info('Moved %s, %s: from %s to %s', str(collection_), 
			str(resource_type), str(src_path), str(dest_path))

	def resourceCreated(self, folder_, src_path):
		'''Event handler for new resources. Creates a background thread
			to upload the resource and add it to the database.'''

		# Retrieve collection and reference to the database
		collection_ = timedrive.findCollection(self.sync_collections, src_path)
		file_localpath = str(src_path)

		# Add description of event to the log
		resource_type = 'folder' if folder_ else 'file'
		logging.info('Created %s, %s: %s', str(collection_), 
			str(resource_type), str(src_path))

		# Check to see if the resource is being manipulated
		if str(src_path) in self.active_ops.keys():
			return False

		# Add resource to active transfers
		self.active_ops[str(src_path)] = sync.NEW

		# Sync local file
		self.checks.append((collection_, sync.LOCAL_CHECK_NEW, file_localpath))
		self.startCheck()
		return True

	def resourceDeleted(self, folder_, src_path):
		'''Event handler for resources that have been deleted. Creates a
			background thread to remove the resource from the remote
			file store.'''
		
		# Retrieve collection and reference to the database
		collection_ = timedrive.findCollection(
			self.sync_collections, src_path)
		file_localpath = str(src_path)
		
		# Add description of event to the log
		resource_type = 'folder' if folder_ else 'file'
		logging.info('Deleted %s, %s: %s', str(collection_),
			str(resource_type), str(src_path))

		# Check to see if the resource is being manipulated
		if str(src_path) in self.active_ops.keys():
			return False

		# Add the resource to the active transfer list, begin operation
		self.active_ops[file_localpath] = sync.DELETE
		self.checks.append((collection_, sync.LOCAL_CHECK_DELETION, file_localpath))
		self.startCheck()
		return True

	def resourceModified(self, folder_, src_path):
		'''Event handler for resources that have been modified. Creates
			a background thread to update the resource and modify the
			entry in the database.'''

		# Retrieve collection and reference to the database
		collection_ = timedrive.findCollection(
			self.sync_collections, src_path)
		file_localpath = str(src_path)

		# Add description of event to the log
		resource_type = 'folder' if folder_ else 'file'
		logging.info('Modified %s, %s: %s', str(collection_), 
			str(resource_type), str(src_path))

		# Check to see if the resource is being manipulated
		if str(src_path) in self.active_ops.keys():
			return False

		# Add the resource to the active transfer list, begin sync
		self.active_ops[file_localpath] = sync.UPDATE
		self.checks.append((collection_, sync.LOCAL_CHECK_EXISTING, file_localpath))
		self.startCheck()
		return True

	def cleanUpSync(self):
		'''Wait for remote operations to complete. Close connections to databases.'''
		logging.info('Exiting Application')
		# Close database sessions
		for collection_ in self.sync_db:
			db_engine, db_session = self.sync_db[collection_]
			db_session.close()


if __name__ == '__main__':

	import sys

	app = QtGui.QApplication(sys.argv)

	if not QtGui.QSystemTrayIcon.isSystemTrayAvailable():
		QtGui.QMessageBox.critical(None, "Systray",
			"I couldn't detect any system tray on this system")
		sys.exit(1)

	QtGui.QApplication.setQuitOnLastWindowClosed(False)

	window = TimeDriveSystemTray()
	window.show()
	sys.exit(app.exec_())