#!/usr/bin/python

import os, sys, datetime, logging, posixpath
from lxml import etree

# Time-Drive Imports
import sync
from sync.local import SyncFile
from sync.backend import SyncBackend

# Script Logic

if __name__ == '__main__':

	# Look at a directory on disk, upload a new version of a document
	# if it differs from the version in the database
	# Differences are calculated by md5 checksum
	# Script checks the beginning and end of the file, rather than calculating
	# a hash for the entire file all at once.

	# Load Application Settings and Sync Collections
	appsettings = sync.appSettings('resources')
	sync_collections = sync.collectionSettings('resources')

	# Setup Backend
	server_ = appsettings['AppSettings']['Scheme'] + '://' \
		+ appsettings['AppSettings']['Host']
	server_url = posixpath.join(server_, appsettings['AppSettings']['Path'])
	sync_backend = SyncBackend(server_url, # Url to Sync To
		appsettings['AppSettings']['Username'], # Username
		appsettings['AppSettings']['Password'], # Password
		appsettings['AppSettings']['Backend']) # Sync Backend to Use
	sync_backend.setExcludeList(appsettings['AppSettings']['Exclude'])
	sync_backend.settingsFolder(appsettings['AppSettings']['Remote-Settings'])

	# Configure Application Log
	ch = logging.StreamHandler()
	log = logging.getLogger()
	slog = logging.getLogger()

	logging.basicConfig(filename = os.path.join(
			appsettings['AppSettings']['Local-Settings'], 
			'Time-Drive.'+datetime.datetime.strftime(
				datetime.datetime.now(), '%Y-%m%d.%H%M')+'.log'),
		)

	log.setLevel(logging.WARNING)

	slog.addHandler(ch)
	slog.setLevel(logging.INFO)

	logging.info('Backup Location: ' + server_url)

	# Start Backup of Collections
	for collection_name in sync_collections.keys():

		# Retrieve Settings for Collection
		sync_collection = sync_collections[collection_name]

		# Create Connection to the Database
		db_path = os.path.join(appsettings['AppSettings']['Local-Settings'], 
			sync_collection['Database'])
		db_engine, db_session = sync.createDataConnection(db_path)

		# Ensure that Database Tables Exist
		sync.createDataTables(db_engine)

		# Retrieve Url for Local Folder
		wd_ = os.path.abspath(sync_collection['Local-Path'])
		rd_ = posixpath.join(server_url, sync_collection['Remote-Path'])

		# Start Backup of the Local Folder
		sync.folderBackup(wd_, rd_, sync_backend, db_session)

		# Generate manifest and upload to the server
		manifest_ = sync.createFullSyncManifest(collection_name, wd_, db_session)
		sync_backend.uploadManifest(collection_name, etree.tostring(manifest_, pretty_print=True))
		logging.info('Manifest Upload: ' + collection_name)

	logging.info('Backup Completed Successfully: ' + str(datetime.datetime.now()))