#!/usr/bin/python

import os, sys, datetime, logging, posixpath
from lxml import etree

# Time-Drive imports
import sync
from sync.local import SyncFile
from sync.backend import SyncBackend

# Script Logic

if __name__ == '__main__':

	# Downloads a manifest file from a remote server which contains information
	# about the files on the server. Looks for differences as compared to the local
	# database. If newer (and different) files are found, it replaces the local files
	# in the collection.

	# Looks at directories on disk for differences to those stored in 
	# the sync database. When a file that is different and newer is located,
	# it uploads the files to the remote server.

	# Differences are calculated by looking at the files mdate and the md5 checksum.

	# Load Application Settings and Sync Collections
	appsettings = sync.appSettings('resources')
	sync_collections = sync.collectionSettings('resources')

	# Setup backend
	server_ = appsettings['AppSettings']['Scheme'] + '://' \
		+ appsettings['AppSettings']['Host']
	server_url = posixpath.join(server_, appsettings['AppSettings']['Path'])
	sync_backend = SyncBackend(server_url, #Url of Server
		appsettings['AppSettings']['Username'], # Username
		appsettings['AppSettings']['Password'], # Password
		appsettings['AppSettings']['Backend']) # Sync Backend to Use
	sync_backend.setExcludeList(appsettings['AppSettings']['Exclude'])
	sync_backend.settingsFolder(appsettings['AppSettings']['Remote-Settings'])

	# Configure Application Log
	ch = logging.StreamHandler()
	log = logging.getLogger()
	slog = logging.getLogger()

	logging.basicConfig(filename = os.path.join(
			appsettings['AppSettings']['Local-Settings'], 
			'Time-Drive.'+datetime.datetime.strftime(
				datetime.datetime.now(), '%Y-%m%d.%H%M')+'.log'),
		)

	log.setLevel(logging.WARNING)

	slog.addHandler(ch)
	slog.setLevel(logging.INFO)

	logging.info('Sync Location: ' + server_url)

	# Start Synchronization of Collection
	for collection_name in sync_collections.keys():
		
		# Retrieve Settings for Collection
		sync_collection = sync_collections[collection_name]

		# Create Connection to the Database
		db_path = os.path.join(appsettings['AppSettings']['Local-Settings'],
			sync_collection['Database'])
		db_engine, db_session = sync.createDataConnection(db_path)

		# Ensure that the database tables Exist
		sync.createDataTables(db_engine)

		# Retrieve Url for Local Folder
		wd_ = os.path.abspath(sync_collection['Local-Path'])
		rd_ = posixpath.join(server_url, sync_collection['Remote-Path'])

		rd_filelist = sync_backend.remoteFileList(collection_name, wd_, rd_)

		if rd_filelist:
			# Synchronize with the directory/database on the local computer
			if (len(rd_filelist) > 0) & (os.path.exists(wd_) is True):
				sync.folderSync(collection_name, wd_, rd_,
					sync_backend, db_session, rd_filelist)
			elif (len(rd_filelist) > 0) & (os.path.exists(wd_) is False):
				logging.info('No files from collection found.')
				logging.info('Downloading files from the server.')
				sync.downloadCollection(wd_, rd_, sync_backend, 
					db_session, rd_filelist)

		# If no manifest, upload all files using a backup operation
		else:
			logging.info('Unable to locate manifest for collection ' + collection_name)
			logging.info('Uploading files to the server')
			sync.folderBackup(wd_, rd_, sync_backend, db_session)

		# Finish Sync Operation
		sync_backend.finishSync(collection_name, wd_, db_session)

	logging.info(
		'Sync Finished Successfully for all Collections: ' + str(datetime.datetime.utcnow()))