from PyQt4 import QtCore
import logging, time
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler
import sync

# Background Operations

class BackgroundTask(QtCore.QThread):
	'''BackgroundTask class: used for long running processes'''

	def __init__(self, taskname, parent=None):
		'''Initialization method for Background Task'''
		QtCore.QThread.__init__(self, parent)
		self.task_name = taskname
		self.settings = None

	def taskComplete(self, taskname):
		'''Emit a message when a run is complete'''
		self.emit(QtCore.SIGNAL('taskComplete(QString)'),
			QtCore.QString(taskname))

	def updateTaskProgress(self, progress, message=''):
		'''Emits an update to the main GUI. Used for progressbar and UI updates.'''
		self.emit(QtCore.SIGNAL('updateTaskProgress(int, QString)'),
			progress, QtCore.QString(message))

	def errorMessage(self, title, message, details=''):
		'''Emits a signal with a specific error code'''
		self.emit(QtCore.SIGNAL('errorMessage(QString, QString, QString)'),
			title, message, details)


class SyncTask(BackgroundTask):
	'''Background threads which communicate with the online provider. Used for
		both checking which action needs to be performed and for performing them.'''
	
	def __init__(self, taskname, method_, args_, parent=None):
		'''Initialize the background task'''

		BackgroundTask.__init__(self, taskname, parent)
		self.method = method_
		self.args = args_
		self.results = None
		self.file_list = None

	def run(self):
		'''Run the background task'''

		# Determine type of task and how results should be handled
		if self.task_name in [sync.LOCAL_CHECK_FOLDER_CHANGES, 
			sync.RT_CHECK_FOLDER_CHANGES]:
			self.results, self.file_list = self.method(*self.args)
		else:
			self.results = self.method(*self.args)
		
		self.taskComplete(self.task_name)


class FolderWatcher(QtCore.QThread):
	'''Background class that watches for changes in files and folders.
		Uses the Pythn Watchdog python class to detect changes.'''

	watchpaths = None
	observers = None

	def run(self):
		'''Start watching folders on the target system.'''

		logging.info('Starting folder watcher')
		self.event_handler = SyncEventsHandler()
		self.event_handler.folderWatcher = self
		self.observers = {}

		# Create watcher objects for each collection
		for collection_ in self.watchpaths:
			collection_localpath = self.watchpaths[collection_]
			observer = Observer()
			observer.schedule(self.event_handler, 
				path=collection_localpath, recursive=True)
			observer.start()
			self.observers[collection_] = observer

		try:
			while True:
				time.sleep(1)
		except:
			observer.stop()
		observer.join()

	def resourceMoved(self, folder_, src_path, dest_path):
		'''Resource moved from one location to another. Signal communicates
			if the resource if a folder, the original path, and the destination.'''
		self.emit(QtCore.SIGNAL('resourceMoved(bool, QString, QString)'),
			folder_, src_path, dest_path)

	def resourceCreated(self, folder_, src_path):
		'''Resource created at the source path.'''
		self.emit(QtCore.SIGNAL('resourceCreated(bool, QString)'),
			folder_, src_path)

	def resourceDeleted(self, folder_, src_path):
		'''Resource deleted at the source path.'''
		self.emit(QtCore.SIGNAL('resourceDeleted(bool, QString)'),
			folder_, src_path)

	def resourceModified(self, folder_, src_path):
		'''Resource modified.'''
		self.emit(QtCore.SIGNAL('resourceModified(bool, QString)'),
			folder_, src_path)


# Control Structures

class SyncEventsHandler(FileSystemEventHandler):
	'''The SyncEventsHandler integrates the Watchdog file events with QtSignals
		and slots that can be used to process them.'''

	folderWatcher = None

	def on_moved (self, event):
		'''Event that fires when a file has been moved'''
		super(SyncEventsHandler, self).on_moved(event)
		self.folderWatcher.resourceMoved(event.is_directory, 
			event.src_path, event.dest_path)

	def on_created(self, event):
		'''Event that fires when a file has been created'''
		super(SyncEventsHandler, self).on_created(event)
		self.folderWatcher.resourceCreated(event.is_directory, event.src_path)

	def on_deleted(self, event):
		'''Event that fires when a file is deleted'''
		super(SyncEventsHandler, self).on_deleted(event)
		self.folderWatcher.resourceDeleted(event.is_directory, event.src_path)

	def on_modified(self, event):
		super(SyncEventsHandler, self).on_modified(event)
		self.folderWatcher.resourceModified(event.is_directory, event.src_path)